"""
Query and modify an LDAP database (alternative interface)
=========================================================

.. versionadded:: 2016.3.0

This is an alternative to the ``ldap`` interface provided by the
:py:mod:`ldapmod <salt.modules.ldapmod>` execution module.

:depends: - ``ldap`` Python module
"""

import logging

import salt.utils.data

available_backends = set()
try:
    import ldap
    import ldap.ldapobject  # pylint: disable=no-name-in-module
    import ldap.modlist  # pylint: disable=no-name-in-module
    import ldap.sasl  # pylint: disable=no-name-in-module

    available_backends.add("ldap")
except ImportError:
    pass


log = logging.getLogger(__name__)


def __virtual__():
    """Only load this module if the Python ldap module is present"""
    if available_backends:
        return True
    return False


class LDAPError(Exception):
    """Base class of all LDAP exceptions raised by backends.

    This is only used for errors encountered while interacting with
    the LDAP server; usage errors (e.g., invalid backend name) will
    have a different type.

    :ivar cause: backend exception object, if applicable
    """

    def __init__(self, message, cause=None):
        super().__init__(message)
        self.cause = cause


def _convert_exception(e):
    """Convert an ldap backend exception to an LDAPError and raise it."""
    raise LDAPError(f"exception in ldap backend: {e!r}", e) from e


def _bind(l, bind=None):
    """Bind helper."""
    if bind is None:
        return
    method = bind.get("method", "simple")
    if method is None:
        return
    elif method == "simple":
        l.simple_bind_s(bind.get("dn", ""), bind.get("password", ""))
    elif method == "sasl":
        sasl_class = getattr(ldap.sasl, bind.get("mechanism", "EXTERNAL").lower())
        creds = bind.get("credentials", None)
        if creds is None:
            creds = {}
        auth = sasl_class(*creds.get("args", []), **creds.get("kwargs", {}))
        l.sasl_interactive_bind_s(bind.get("dn", ""), auth)
    else:
        raise ValueError(
            'unsupported bind method "'
            + method
            + '"; supported bind methods: simple sasl'
        )


def _format_unicode_password(pwd):
    """Formats a string per Microsoft AD password specifications.
    The string must be enclosed in double quotes and UTF-16 encoded.
    See: https://msdn.microsoft.com/en-us/library/cc223248.aspx

    :param pwd:
       The desired password as a string

    :returns:
        A unicode string
    """
    return f'"{pwd}"'.encode("utf-16-le")


class _connect_ctx:
    def __init__(self, c):
        self.c = c

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        pass


def connect(connect_spec=None):
    """Connect and optionally bind to an LDAP server.

    :param connect_spec:
        This can be an LDAP connection object returned by a previous
        call to :py:func:`connect` (in which case the argument is
        simply returned), ``None`` (in which case an empty dict is
        used), or a dict with the following keys:

        * ``'backend'``
            Optional; default depends on which Python LDAP modules are
            installed.  Name of the Python LDAP module to use.  Only
            ``'ldap'`` is supported at the moment.

        * ``'url'``
            Optional; defaults to ``'ldapi:///'``.  URL to the LDAP
            server.

        * ``'bind'``
            Optional; defaults to ``None``.  Describes how to bind an
            identity to the LDAP connection.  If ``None``, an
            anonymous connection is made.  Valid keys:

            * ``'method'``
                Optional; defaults to ``None``.  The authentication
                method to use.  Valid values include but are not
                necessarily limited to ``'simple'``, ``'sasl'``, and
                ``None``.  If ``None``, an anonymous connection is
                made.  Available methods depend on the chosen backend.

            * ``'mechanism'``
                Optional; defaults to ``'EXTERNAL'``.  The SASL
                mechanism to use.  Ignored unless the method is
                ``'sasl'``.  Available methods depend on the chosen
                backend and the server's capabilities.

            * ``'credentials'``
                Optional; defaults to ``None``.  An object specific to
                the chosen SASL mechanism and backend that represents
                the authentication credentials.  Ignored unless the
                method is ``'sasl'``.

                For the ``'ldap'`` backend, this is a dictionary.  If
                ``None``, an empty dict is used.  Keys:

                * ``'args'``
                    Optional; defaults to an empty list.  A list of
                    arguments to pass to the SASL mechanism
                    constructor.  See the SASL mechanism constructor
                    documentation in the ``ldap.sasl`` Python module.

                * ``'kwargs'``
                    Optional; defaults to an empty dict.  A dict of
                    keyword arguments to pass to the SASL mechanism
                    constructor.  See the SASL mechanism constructor
                    documentation in the ``ldap.sasl`` Python module.

            * ``'dn'``
                Optional; defaults to an empty string.  The
                distinguished name to bind.

            * ``'password'``
                Optional; defaults to an empty string.  Password for
                binding.  Ignored if the method is ``'sasl'``.

        * ``'tls'``
            Optional; defaults to ``None``.  A backend-specific object
            containing settings to override default TLS behavior.

            For the ``'ldap'`` backend, this is a dictionary.  Not all
            settings in this dictionary are supported by all versions
            of ``python-ldap`` or the underlying TLS library.  If
            ``None``, an empty dict is used.  Possible keys:

            * ``'starttls'``
                If present, initiate a TLS connection using StartTLS.
                (The value associated with this key is ignored.)

            * ``'cacertdir'``
                Set the path of the directory containing CA
                certificates.

            * ``'cacertfile'``
                Set the pathname of the CA certificate file.

            * ``'certfile'``
                Set the pathname of the certificate file.

            * ``'cipher_suite'``
                Set the allowed cipher suite.

            * ``'crlcheck'``
                Set the CRL evaluation strategy.  Valid values are
                ``'none'``, ``'peer'``, and ``'all'``.

            * ``'crlfile'``
                Set the pathname of the CRL file.

            * ``'dhfile'``
                Set the pathname of the file containing the parameters
                for Diffie-Hellman ephemeral key exchange.

            * ``'keyfile'``
                Set the pathname of the certificate key file.

            * ``'newctx'``
                If present, instruct the underlying TLS library to
                create a new TLS context.  (The value associated with
                this key is ignored.)

            * ``'protocol_min'``
                Set the minimum protocol version.

            * ``'random_file'``
                Set the pathname of the random file when
                ``/dev/random`` and ``/dev/urandom`` are not
                available.

            * ``'require_cert'``
                Set the certificate validation policy.  Valid values
                are ``'never'``, ``'hard'``, ``'demand'``,
                ``'allow'``, and ``'try'``.

        * ``'opts'``
            Optional; defaults to ``None``.  A backend-specific object
            containing options for the backend.

            For the ``'ldap'`` backend, this is a dictionary of
            OpenLDAP options to set.  If ``None``, an empty dict is
            used.  Each key is a the name of an OpenLDAP option
            constant without the ``'LDAP_OPT_'`` prefix, then
            converted to lower case.

    :returns:
        an object representing an LDAP connection that can be used as
        the ``connect_spec`` argument to any of the functions in this
        module (to avoid the overhead of making and terminating
        multiple connections).

        This object should be used as a context manager.  It is safe
        to nest ``with`` statements.

    CLI Example:

    .. code-block:: bash

        salt '*' ldap3.connect "{
            'url': 'ldaps://ldap.example.com/',
            'bind': {
                'method': 'simple',
                'dn': 'cn=admin,dc=example,dc=com',
                'password': 'secret'}
        }"
    """
    if isinstance(connect_spec, _connect_ctx):
        return connect_spec
    if connect_spec is None:
        connect_spec = {}
    backend_name = connect_spec.get("backend", "ldap")
    if backend_name not in available_backends:
        raise ValueError(
            "unsupported backend or required Python module"
            + f" unavailable: {backend_name}"
        )
    url = connect_spec.get("url", "ldapi:///")
    try:
        l = ldap.initialize(url)
        l.protocol_version = ldap.VERSION3

        # set up tls
        tls = connect_spec.get("tls", None)
        if tls is None:
            tls = {}
        vars = {}
        for k, v in tls.items():
            if k in ("starttls", "newctx"):
                vars[k] = True
            elif k in ("crlcheck", "require_cert"):
                l.set_option(
                    getattr(ldap, "OPT_X_TLS_" + k.upper()),
                    getattr(ldap, "OPT_X_TLS_" + v.upper()),
                )
            else:
                l.set_option(getattr(ldap, "OPT_X_TLS_" + k.upper()), v)
        if vars.get("starttls", False):
            l.start_tls_s()
        if vars.get("newctx", False):
            l.set_option(ldap.OPT_X_TLS_NEWCTX, 0)

        # set up other options
        l.set_option(ldap.OPT_REFERRALS, 0)
        opts = connect_spec.get("opts", None)
        if opts is None:
            opts = {}
        for k, v in opts.items():
            opt = getattr(ldap, "OPT_" + k.upper())
            l.set_option(opt, v)

        _bind(l, connect_spec.get("bind", None))
    except ldap.LDAPError as e:
        _convert_exception(e)
    return _connect_ctx(l)


def search(
    connect_spec,
    base,
    scope="subtree",
    filterstr="(objectClass=*)",
    attrlist=None,
    attrsonly=0,
):
    """Search an LDAP database.

    :param connect_spec:
        See the documentation for the ``connect_spec`` parameter for
        :py:func:`connect`.

    :param base:
        Distinguished name of the entry at which to start the search.

    :param scope:
        One of the following:

        * ``'subtree'``
            Search the base and all of its descendants.

        * ``'base'``
            Search only the base itself.

        * ``'onelevel'``
            Search only the base's immediate children.

    :param filterstr:
        String representation of the filter to apply in the search.

    :param attrlist:
        Limit the returned attributes to those in the specified list.
        If ``None``, all attributes of each entry are returned.

    :param attrsonly:
        If non-zero, don't return any attribute values.

    :returns:
        a dict of results.  The dict is empty if there are no results.
        The dict maps each returned entry's distinguished name to a
        dict that maps each of the matching attribute names to a list
        of its values.

    CLI Example:

    .. code-block:: bash

        salt '*' ldap3.search "{
            'url': 'ldaps://ldap.example.com/',
            'bind': {
                'method': 'simple',
                'dn': 'cn=admin,dc=example,dc=com',
                'password': 'secret',
            },
        }" "base='dc=example,dc=com'"
    """
    l = connect(connect_spec)
    scope = getattr(ldap, "SCOPE_" + scope.upper())
    try:
        results = l.c.search_s(base, scope, filterstr, attrlist, attrsonly)
    except ldap.NO_SUCH_OBJECT:
        results = []
    except ldap.LDAPError as e:
        _convert_exception(e)
    return dict(results)


def add(connect_spec, dn, attributes):
    """Add an entry to an LDAP database.

    :param connect_spec:
        See the documentation for the ``connect_spec`` parameter for
        :py:func:`connect`.

    :param dn:
        Distinguished name of the entry.

    :param attributes:
        Non-empty dict mapping each of the new entry's attributes to a
        non-empty iterable of values.

    :returns:
        ``True`` if successful, raises an exception otherwise.

    CLI Example:

    .. code-block:: bash

        salt '*' ldap3.add "{
            'url': 'ldaps://ldap.example.com/',
            'bind': {
                'method': 'simple',
                'password': 'secret',
            },
        }" "dn='dc=example,dc=com'" "attributes={'example': 'values'}"
    """
    l = connect(connect_spec)
    # convert the "iterable of values" to lists in case that's what
    # addModlist() expects (also to ensure that the caller's objects
    # are not modified)
    attributes = {
        attr: salt.utils.data.encode(list(vals)) for attr, vals in attributes.items()
    }
    log.info("adding entry: dn: %s attributes: %s", repr(dn), repr(attributes))

    if "unicodePwd" in attributes:
        attributes["unicodePwd"] = [
            _format_unicode_password(x) for x in attributes["unicodePwd"]
        ]

    modlist = ldap.modlist.addModlist(attributes)
    try:
        l.c.add_s(dn, modlist)
    except ldap.LDAPError as e:
        _convert_exception(e)
    return True


def delete(connect_spec, dn):
    """Delete an entry from an LDAP database.

    :param connect_spec:
        See the documentation for the ``connect_spec`` parameter for
        :py:func:`connect`.

    :param dn:
        Distinguished name of the entry.

    :returns:
        ``True`` if successful, raises an exception otherwise.

    CLI Example:

    .. code-block:: bash

        salt '*' ldap3.delete "{
            'url': 'ldaps://ldap.example.com/',
            'bind': {
                'method': 'simple',
                'password': 'secret'}
        }" dn='cn=admin,dc=example,dc=com'
    """
    l = connect(connect_spec)
    log.info("deleting entry: dn: %s", repr(dn))
    try:
        l.c.delete_s(dn)
    except ldap.LDAPError as e:
        _convert_exception(e)
    return True


def modify(connect_spec, dn, directives):
    """Modify an entry in an LDAP database.

    :param connect_spec:
        See the documentation for the ``connect_spec`` parameter for
        :py:func:`connect`.

    :param dn:
        Distinguished name of the entry.

    :param directives:
        Iterable of directives that indicate how to modify the entry.
        Each directive is a tuple of the form ``(op, attr, vals)``,
        where:

        * ``op`` identifies the modification operation to perform.
          One of:

          * ``'add'`` to add one or more values to the attribute

          * ``'delete'`` to delete some or all of the values from the
            attribute.  If no values are specified with this
            operation, all of the attribute's values are deleted.
            Otherwise, only the named values are deleted.

          * ``'replace'`` to replace all of the attribute's values
            with zero or more new values

        * ``attr`` names the attribute to modify

        * ``vals`` is an iterable of values to add or delete

    :returns:
        ``True`` if successful, raises an exception otherwise.

    CLI Example:

    .. code-block:: bash

        salt '*' ldap3.modify "{
            'url': 'ldaps://ldap.example.com/',
            'bind': {
                'method': 'simple',
                'password': 'secret'}
        }" dn='cn=admin,dc=example,dc=com'
        directives="('add', 'example', ['example_val'])"
    """
    l = connect(connect_spec)
    # convert the "iterable of values" to lists in case that's what
    # modify_s() expects (also to ensure that the caller's objects are
    # not modified)
    modlist = [
        (getattr(ldap, "MOD_" + op.upper()), attr, list(vals))
        for op, attr, vals in directives
    ]

    for idx, mod in enumerate(modlist):
        if mod[1] == "unicodePwd":
            modlist[idx] = (
                mod[0],
                mod[1],
                [_format_unicode_password(x) for x in mod[2]],
            )

    modlist = salt.utils.data.decode(modlist, to_str=True, preserve_tuples=True)
    try:
        l.c.modify_s(dn, modlist)
    except ldap.LDAPError as e:
        _convert_exception(e)
    return True


def change(connect_spec, dn, before, after):
    """Modify an entry in an LDAP database.

    This does the same thing as :py:func:`modify`, but with a simpler
    interface.  Instead of taking a list of directives, it takes a
    before and after view of an entry, determines the differences
    between the two, computes the directives, and executes them.

    Any attribute value present in ``before`` but missing in ``after``
    is deleted.  Any attribute value present in ``after`` but missing
    in ``before`` is added.  Any attribute value in the database that
    is not mentioned in either ``before`` or ``after`` is not altered.
    Any attribute value that is present in both ``before`` and
    ``after`` is ignored, regardless of whether that attribute value
    exists in the database.

    :param connect_spec:
        See the documentation for the ``connect_spec`` parameter for
        :py:func:`connect`.

    :param dn:
        Distinguished name of the entry.

    :param before:
        The expected state of the entry before modification.  This is
        a dict mapping each attribute name to an iterable of values.

    :param after:
        The desired state of the entry after modification.  This is a
        dict mapping each attribute name to an iterable of values.

    :returns:
        ``True`` if successful, raises an exception otherwise.

    CLI Example:

    .. code-block:: bash

        salt '*' ldap3.change "{
            'url': 'ldaps://ldap.example.com/',
            'bind': {
                'method': 'simple',
                'password': 'secret'}
        }" dn='cn=admin,dc=example,dc=com'
        before="{'example_value': 'before_val'}"
        after="{'example_value': 'after_val'}"
    """
    l = connect(connect_spec)
    # convert the "iterable of values" to lists in case that's what
    # modifyModlist() expects (also to ensure that the caller's dicts
    # are not modified)
    before = {attr: salt.utils.data.encode(list(vals)) for attr, vals in before.items()}
    after = {attr: salt.utils.data.encode(list(vals)) for attr, vals in after.items()}

    if "unicodePwd" in after:
        after["unicodePwd"] = [_format_unicode_password(x) for x in after["unicodePwd"]]

    modlist = ldap.modlist.modifyModlist(before, after)

    try:
        l.c.modify_s(dn, modlist)
    except ldap.LDAPError as e:
        _convert_exception(e)
    return True



"""
Used to manage the thorium register. The thorium register is where compound
values are stored and computed, such as averages etc.
"""

import salt.utils.stringutils

__func_alias__ = {
    "set_": "set",
    "list_": "list",
}


def set_(name, add, match):
    """
    Add a value to the named set

    USAGE:

    .. code-block:: yaml

        foo:
          reg.set:
            - add: bar
            - match: my/custom/event
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    if name not in __reg__:
        __reg__[name] = {}
        __reg__[name]["val"] = set()
    for event in __events__:
        if salt.utils.stringutils.expr_match(event["tag"], match):
            try:
                val = event["data"]["data"].get(add)
            except KeyError:
                val = event["data"].get(add)
            if val is None:
                val = "None"
            ret["changes"][add] = val
            __reg__[name]["val"].add(val)
    return ret


def list_(name, add, match, stamp=False, prune=0):
    """
    Add the specified values to the named list

    If ``stamp`` is True, then the timestamp from the event will also be added
    if ``prune`` is set to an integer higher than ``0``, then only the last
    ``prune`` values will be kept in the list.

    USAGE:

    .. code-block:: yaml

        foo:
          reg.list:
            - add: bar
            - match: my/custom/event
            - stamp: True
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    if not isinstance(add, list):
        add = add.split(",")
    if name not in __reg__:
        __reg__[name] = {}
        __reg__[name]["val"] = []
    for event in __events__:
        try:
            event_data = event["data"]["data"]
        except KeyError:
            event_data = event["data"]
        if salt.utils.stringutils.expr_match(event["tag"], match):
            item = {}
            for key in add:
                if key in event_data:
                    item[key] = event_data[key]
                    if stamp is True:
                        item["time"] = event["data"]["_stamp"]
            __reg__[name]["val"].append(item)
    if prune > 0:
        __reg__[name]["val"] = __reg__[name]["val"][:prune]
    return ret


def mean(name, add, match):
    """
    Accept a numeric value from the matched events and store a running average
    of the values in the given register. If the specified value is not numeric
    it will be skipped

    USAGE:

    .. code-block:: yaml

        foo:
          reg.mean:
            - add: data_field
            - match: my/custom/event
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    if name not in __reg__:
        __reg__[name] = {}
        __reg__[name]["val"] = 0
        __reg__[name]["total"] = 0
        __reg__[name]["count"] = 0
    for event in __events__:
        try:
            event_data = event["data"]["data"]
        except KeyError:
            event_data = event["data"]
        if salt.utils.stringutils.expr_match(event["tag"], match):
            if add in event_data:
                try:
                    comp = int(event_data)
                except ValueError:
                    continue
            __reg__[name]["total"] += comp
            __reg__[name]["count"] += 1
            __reg__[name]["val"] = __reg__[name]["total"] / __reg__[name]["count"]
    return ret


def clear(name):
    """
    Clear the namespace from the register

    USAGE:

    .. code-block:: yaml

        clearns:
          reg.clear:
            - name: myregister
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    if name in __reg__:
        __reg__[name].clear()
    return ret


def delete(name):
    """
    Delete the namespace from the register

    USAGE:

    .. code-block:: yaml

        deletens:
          reg.delete:
            - name: myregister
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    if name in __reg__:
        del __reg__[name]
    return ret



"""
Glance module for interacting with OpenStack Glance

.. versionadded:: 2018.3.0

:depends:shade

Example configuration

.. code-block:: yaml

    glance:
      cloud: default

.. code-block:: yaml

    glance:
      auth:
        username: admin
        password: password123
        user_domain_name: mydomain
        project_name: myproject
        project_domain_name: myproject
        auth_url: https://example.org:5000/v3
      identity_api_version: 3
"""

HAS_SHADE = False
try:
    import shade

    HAS_SHADE = True
except ImportError:
    pass

__virtualname__ = "glanceng"


def __virtual__():
    """
    Only load this module if shade python module is installed
    """
    if HAS_SHADE:
        return __virtualname__
    return (
        False,
        "The glanceng execution module failed to load: shade python module is not"
        " available",
    )


def compare_changes(obj, **kwargs):
    """
    Compare two dicts returning only keys that exist in the first dict and are
    different in the second one
    """
    changes = {}
    for k, v in obj.items():
        if k in kwargs:
            if v != kwargs[k]:
                changes[k] = kwargs[k]
    return changes


def _clean_kwargs(keep_name=False, **kwargs):
    """
    Sanatize the arguments for use with shade
    """
    if "name" in kwargs and not keep_name:
        kwargs["name_or_id"] = kwargs.pop("name")

    return __utils__["args.clean_kwargs"](**kwargs)


def setup_clouds(auth=None):
    """
    Call functions to create Shade cloud objects in __context__ to take
    advantage of Shade's in-memory caching across several states
    """
    get_operator_cloud(auth)
    get_openstack_cloud(auth)


def get_operator_cloud(auth=None):
    """
    Return an operator_cloud
    """
    if auth is None:
        auth = __salt__["config.option"]("glance", {})
    if "shade_opcloud" in __context__:
        if __context__["shade_opcloud"].auth == auth:
            return __context__["shade_opcloud"]
    __context__["shade_opcloud"] = shade.operator_cloud(**auth)
    return __context__["shade_opcloud"]


def get_openstack_cloud(auth=None):
    """
    Return an openstack_cloud
    """
    if auth is None:
        auth = __salt__["config.option"]("glance", {})
    if "shade_oscloud" in __context__:
        if __context__["shade_oscloud"].auth == auth:
            return __context__["shade_oscloud"]
    __context__["shade_oscloud"] = shade.openstack_cloud(**auth)
    return __context__["shade_oscloud"]


def image_create(auth=None, **kwargs):
    """
    Create an image

    CLI Example:

    .. code-block:: bash

        salt '*' glanceng.image_create name=cirros file=cirros.raw disk_format=raw
        salt '*' glanceng.image_create name=cirros file=cirros.raw disk_format=raw hw_scsi_model=virtio-scsi hw_disk_bus=scsi
    """
    cloud = get_operator_cloud(auth)
    kwargs = _clean_kwargs(keep_name=True, **kwargs)
    return cloud.create_image(**kwargs)


def image_delete(auth=None, **kwargs):
    """
    Delete an image

    CLI Example:

    .. code-block:: bash

        salt '*' glanceng.image_delete name=image1
        salt '*' glanceng.image_delete name=0e4febc2a5ab4f2c8f374b054162506d
    """
    cloud = get_operator_cloud(auth)
    kwargs = _clean_kwargs(**kwargs)
    return cloud.delete_image(**kwargs)


def image_list(auth=None, **kwargs):
    """
    List images

    CLI Example:

    .. code-block:: bash

        salt '*' glanceng.image_list
        salt '*' glanceng.image_list
    """
    cloud = get_operator_cloud(auth)
    kwargs = _clean_kwargs(**kwargs)
    return cloud.list_images(**kwargs)


def image_search(auth=None, **kwargs):
    """
    Search for images

    CLI Example:

    .. code-block:: bash

        salt '*' glanceng.image_search name=image1
        salt '*' glanceng.image_search
    """
    cloud = get_operator_cloud(auth)
    kwargs = _clean_kwargs(**kwargs)
    return cloud.search_images(**kwargs)


def image_get(auth=None, **kwargs):
    """
    Get a single image

    CLI Example:

    .. code-block:: bash

        salt '*' glanceng.image_get name=image1
        salt '*' glanceng.image_get name=0e4febc2a5ab4f2c8f374b054162506d
    """
    cloud = get_operator_cloud(auth)
    kwargs = _clean_kwargs(**kwargs)
    return cloud.get_image(**kwargs)


def update_image_properties(auth=None, **kwargs):
    """
    Update properties for an image

    CLI Example:

    .. code-block:: bash

        salt '*' glanceng.update_image_properties name=image1 hw_scsi_model=virtio-scsi hw_disk_bus=scsi
        salt '*' glanceng.update_image_properties name=0e4febc2a5ab4f2c8f374b054162506d min_ram=1024
    """
    cloud = get_operator_cloud(auth)
    kwargs = _clean_kwargs(**kwargs)
    return cloud.update_image_properties(**kwargs)



"""
Simple and flexible YAML ext_pillar which can read pillar from within pillar.

.. versionadded:: 2016.3.0

`PillarStack <https://github.com/bbinet/pillarstack>`_ is a custom saltstack
``ext_pillar`` which was inspired by `varstack
<https://github.com/conversis/varstack>`_ but is heavily based on Jinja2 for
maximum flexibility.

It supports the following features:

- multiple config files that are jinja2 templates with support for ``pillar``,
  ``__grains__``, ``__salt__``, ``__opts__`` objects and ``pillarenv``
- a config file renders as an ordered list of files (paths of these files are
  relative to the current config file)
- this list of files are read in ordered as jinja2 templates with support for
  ``stack``, ``pillar``, ``__grains__``, ``__salt__``, ``__opts__`` objects and
  ``pillarenv``
- all these rendered files are then parsed as ``yaml``
- then all yaml dicts are merged in order with support for the following
  merging strategies: ``merge-first``, ``merge-last``, ``remove``, and
  ``overwrite``
- stack config files can be matched based on ``pillar``, ``grains``, or
  ``opts`` values, which make it possible to support kind of self-contained
  environments

Installation
------------

PillarStack is already bundled with Salt since 2016.3.0 version so there is
nothing to install from version 2016.3.0.

If you use an older Salt version or you want to override PillarStack with a
more recent one, follow the installation procedure below.

Installing the PillarStack ``ext_pillar`` is as simple as dropping the
``stack.py`` file in the ``<extension_modules>/pillar`` directory (no external
python module required), given that ``extension_modules`` is set in your
salt-master configuration, see:
https://docs.saltproject.io/en/latest/ref/configuration/master.html#extension-modules

Configuration in Salt
---------------------

Like any other external pillar, its configuration takes place through the
``ext_pillar`` key in the master config file.

However, you can configure PillarStack in 3 different ways:

Single config file
~~~~~~~~~~~~~~~~~~

This is the simplest option, you just need to set the path to your single
PillarStack config file like below:

.. code:: yaml

    ext_pillar:
      - stack: /path/to/stack.cfg

List of config files
~~~~~~~~~~~~~~~~~~~~

You can also provide a list of config files:

.. code:: yaml

    ext_pillar:
      - stack:
          - /path/to/stack1.cfg
          - /path/to/stack2.cfg

Select config files through grains|pillar|opts matching
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

You can also opt for a much more flexible configuration: PillarStack allows one
to select the config files for the current minion based on matching values from
either grains, or pillar, or opts objects.

Here is an example of such a configuration, which should speak by itself:

.. code:: yaml

    ext_pillar:
      - stack:
          pillar:something:
            bar: /path/to/bar/stack.cfg
            foo: /path/to/foo/stack.cfg
          grains:custom:grain:
            value:
              - /path/to/stack1.cfg
              - /path/to/stack2.cfg
          opts:custom:opt:
            value: /path/to/stack0.cfg
          opts:saltenv:
            dev: /path/to/dev/stack.cfg
            __env__: /path/to/__env__/stack.cfg

PillarStack configuration files
-------------------------------

The config files that are referenced in the above ``ext_pillar`` configuration
are jinja2 templates which must render as a simple ordered list of ``yaml``
files that will then be merged to build pillar data.

The path of these ``yaml`` files must be relative to the directory of the
PillarStack config file. These paths support unix style pathname pattern
expansion through the
`Python glob module <https://docs.python.org/2/library/glob.html>`.

The following variables are available in jinja2 templating of PillarStack
configuration files:

- ``pillar``: the pillar data (as passed by Salt to our ``ext_pillar``
  function)
- ``minion_id``: the minion id ;-)
- ``__opts__``: a dictionary of mostly Salt configuration options
- ``__grains__``: a dictionary of the grains of the minion making this pillar
  call
- ``__salt__``: a dictionary of Salt module functions, useful so you don't have
  to duplicate functions that already exist (note: runs on the master)

So you can use all the power of jinja2 to build your list of ``yaml`` files
that will be merged in pillar data.

For example, you could have a PillarStack config file which looks like:

.. code:: jinja

    $ cat /path/to/stack/config.cfg
    core.yml
    common/*.yml
    osarchs/{{ __grains__['osarch'] }}.yml
    oscodenames/{{ __grains__['oscodename'] }}.yml
    {%- for role in pillar.get('roles', []) %}
    roles/{{ role }}.yml
    {%- endfor %}
    minions/{{ minion_id }}.yml

And the whole directory structure could look like:

.. code::

    $ tree /path/to/stack/
    /path/to/stack/
    ├── config.cfg
    ├── core.yml
    ├── common/
    │   ├── xxx.yml
    │   └── yyy.yml
    ├── osarchs/
    │   ├── amd64.yml
    │   └── armhf.yml
    ├── oscodenames/
    │   ├── wheezy.yml
    │   └── jessie.yml
    ├── roles/
    │   ├── web.yml
    │   └── db.yml
    └── minions/
        ├── test-1-dev.yml
        └── test-2-dev.yml

Overall process
---------------

In the above PillarStack configuration, given that test-1-dev minion is an
amd64 platform running Debian Jessie, and which pillar ``roles`` is ``["db"]``,
the following ``yaml`` files would be merged in order:

- ``core.yml``
- ``common/xxx.yml``
- ``common/yyy.yml``
- ``osarchs/amd64.yml``
- ``oscodenames/jessie.yml``
- ``roles/db.yml``
- ``minions/test-1-dev.yml``

Before merging, every files above will be preprocessed as Jinja2 templates.
The following variables are available in Jinja2 templating of ``yaml`` files:

- ``stack``: the PillarStack pillar data object that has currently been merged
  (data from previous ``yaml`` files in PillarStack configuration)
- ``pillar``: the pillar data (as passed by Salt to our ``ext_pillar``
  function)
- ``minion_id``: the minion id ;-)
- ``__opts__``: a dictionary of mostly Salt configuration options
- ``__grains__``: a dictionary of the grains of the minion making this pillar
  call
- ``__salt__``: a dictionary of Salt module functions, useful so you don't have
  to duplicate functions that already exist (note: runs on the master)

So you can use all the power of jinja2 to build your pillar data, and even use
other pillar values that has already been merged by PillarStack (from previous
``yaml`` files in PillarStack configuration) through the ``stack`` variable.

Once a ``yaml`` file has been preprocessed by Jinja2, we obtain a Python dict -
let's call it ``yml_data`` - then, PillarStack will merge this ``yml_data``
dict in the main ``stack`` dict (which contains already merged PillarStack
pillar data).
By default, PillarStack will deeply merge ``yml_data`` in ``stack`` (similarly
to the ``recurse`` salt ``pillar_source_merging_strategy``), but 3 merging
strategies are currently available for you to choose (see next section).

Once every ``yaml`` files have been processed, the ``stack`` dict will contain
your whole own pillar data, merged in order by PillarStack.
So PillarStack ``ext_pillar`` returns the ``stack`` dict, the contents of which
Salt takes care to merge in with all of the other pillars and finally return
the whole pillar to the minion.

Merging strategies
------------------

The way the data from a new ``yaml_data`` dict is merged with the existing
``stack`` data can be controlled by specifying a merging strategy. Right now
this strategy can either be ``merge-last`` (the default), ``merge-first``,
``remove``, or ``overwrite``.

Note that scalar values like strings, integers, booleans, etc. are always
evaluated using the ``overwrite`` strategy (other strategies don't make sense
in that case).

The merging strategy can be set by including a dict in the form of:

.. code:: yaml

    __: <merging strategy>

as the first item of the dict or list.
This allows fine grained control over the merging process.

``merge-last`` (default) strategy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

If the ``merge-last`` strategy is selected (the default), then content of dict
or list variables is merged recursively with previous definitions of this
variable (similarly to the ``recurse`` salt
``pillar_source_merging_strategy``).
This allows for extending previously defined data.

``merge-first`` strategy
~~~~~~~~~~~~~~~~~~~~~~~~

If the ``merge-first`` strategy is selected, then the content of dict or list
variables are swapped between the ``yaml_data`` and ``stack`` objects before
being merged recursively with the ``merge-last`` previous strategy.

``remove`` strategy
~~~~~~~~~~~~~~~~~~~

If the ``remove`` strategy is selected, then content of dict or list variables
in ``stack`` are removed only if the corresponding item is present in the
``yaml_data`` dict.
This allows for removing items from previously defined data.

``overwrite`` strategy
~~~~~~~~~~~~~~~~~~~~~~

If the ``overwrite`` strategy is selected, then the content of dict or list
variables in ``stack`` is overwritten by the content of ``yaml_data`` dict.
So this allows one to overwrite variables from previous definitions.

Merging examples
----------------

Let's go through small examples that should clarify what's going on when a
``yaml_data`` dict is merged in the ``stack`` dict.

When you don't specify any strategy, the default ``merge-last`` strategy is
selected:

+----------------------+-----------------------+-------------------------+
| ``stack``            | ``yaml_data``         | ``stack`` (after merge) |
+======================+=======================+=========================+
| .. code:: yaml       | .. code:: yaml        | .. code:: yaml          |
|                      |                       |                         |
|     users:           |     users:            |     users:              |
|       tom:           |       tom:            |       tom:              |
|         uid: 500     |         uid: 1000     |         uid: 1000       |
|         roles:       |         roles:        |         roles:          |
|           - sysadmin |           - developer |           - sysadmin    |
|       root:          |       mat:            |           - developer   |
|         uid: 0       |         uid: 1001     |       mat:              |
|                      |                       |         uid: 1001       |
|                      |                       |       root:             |
|                      |                       |         uid: 0          |
+----------------------+-----------------------+-------------------------+

Then you can select a custom merging strategy using the ``__`` key in a dict:

+----------------------+-----------------------+-------------------------+
| ``stack``            | ``yaml_data``         | ``stack`` (after merge) |
+======================+=======================+=========================+
| .. code:: yaml       | .. code:: yaml        | .. code:: yaml          |
|                      |                       |                         |
|     users:           |     users:            |     users:              |
|       tom:           |       __: merge-last  |       tom:              |
|         uid: 500     |       tom:            |         uid: 1000       |
|         roles:       |         uid: 1000     |         roles:          |
|           - sysadmin |         roles:        |           - sysadmin    |
|       root:          |           - developer |           - developer   |
|         uid: 0       |       mat:            |       mat:              |
|                      |         uid: 1001     |         uid: 1001       |
|                      |                       |       root:             |
|                      |                       |         uid: 0          |
+----------------------+-----------------------+-------------------------+
| .. code:: yaml       | .. code:: yaml        | .. code:: yaml          |
|                      |                       |                         |
|     users:           |     users:            |     users:              |
|       tom:           |       __: merge-first |       tom:              |
|         uid: 500     |       tom:            |         uid: 500        |
|         roles:       |         uid: 1000     |         roles:          |
|           - sysadmin |         roles:        |           - developer   |
|       root:          |           - developer |           - sysadmin    |
|         uid: 0       |       mat:            |       mat:              |
|                      |         uid: 1001     |         uid: 1001       |
|                      |                       |       root:             |
|                      |                       |         uid: 0          |
+----------------------+-----------------------+-------------------------+
| .. code:: yaml       | .. code:: yaml        | .. code:: yaml          |
|                      |                       |                         |
|     users:           |     users:            |     users:              |
|       tom:           |       __: remove      |       root:             |
|         uid: 500     |       tom:            |         uid: 0          |
|         roles:       |       mat:            |                         |
|           - sysadmin |                       |                         |
|       root:          |                       |                         |
|         uid: 0       |                       |                         |
+----------------------+-----------------------+-------------------------+
| .. code:: yaml       | .. code:: yaml        | .. code:: yaml          |
|                      |                       |                         |
|     users:           |     users:            |     users:              |
|       tom:           |       __: overwrite   |       tom:              |
|         uid: 500     |       tom:            |         uid: 1000       |
|         roles:       |         uid: 1000     |         roles:          |
|           - sysadmin |         roles:        |           - developer   |
|       root:          |           - developer |       mat:              |
|         uid: 0       |       mat:            |         uid: 1001       |
|                      |         uid: 1001     |                         |
+----------------------+-----------------------+-------------------------+

You can also select a custom merging strategy using a ``__`` object in a list:

+----------------+-------------------------+-------------------------+
| ``stack``      | ``yaml_data``           | ``stack`` (after merge) |
+================+=========================+=========================+
| .. code:: yaml | .. code:: yaml          | .. code:: yaml          |
|                |                         |                         |
|     users:     |     users:              |     users:              |
|       - tom    |       - __: merge-last  |       - tom             |
|       - root   |       - mat             |       - root            |
|                |                         |       - mat             |
+----------------+-------------------------+-------------------------+
| .. code:: yaml | .. code:: yaml          | .. code:: yaml          |
|                |                         |                         |
|     users:     |     users:              |     users:              |
|       - tom    |       - __: merge-first |       - mat             |
|       - root   |       - mat             |       - tom             |
|                |                         |       - root            |
+----------------+-------------------------+-------------------------+
| .. code:: yaml | .. code:: yaml          | .. code:: yaml          |
|                |                         |                         |
|     users:     |     users:              |     users:              |
|       - tom    |       - __: remove      |       - root            |
|       - root   |       - mat             |                         |
|                |       - tom             |                         |
+----------------+-------------------------+-------------------------+
| .. code:: yaml | .. code:: yaml          | .. code:: yaml          |
|                |                         |                         |
|     users:     |     users:              |     users:              |
|       - tom    |       - __: overwrite   |       - mat             |
|       - root   |       - mat             |                         |
+----------------+-------------------------+-------------------------+
"""

import functools
import glob
import logging
import os
import posixpath

from jinja2 import Environment, FileSystemLoader

import salt.utils.data
import salt.utils.jinja
import salt.utils.yaml

log = logging.getLogger(__name__)
strategies = ("overwrite", "merge-first", "merge-last", "remove")


def ext_pillar(minion_id, pillar, *args, **kwargs):
    """
    Builds stacked pillar from yaml files listed in file(s).

    :param str minion_id: Minion ID
    :param dict pillar: pillar
    :param list args: (Optional) file(s) that list yaml files
    :param dict kwargs: (Optional) conditional file(s) that list yaml files
    """
    stack = {}
    stack_config_files = list(args)
    pillarenv = __opts__["pillarenv"] or __opts__["saltenv"] or "base"
    traverse = {
        "pillar": functools.partial(salt.utils.data.traverse_dict_and_list, pillar),
        "grains": functools.partial(salt.utils.data.traverse_dict_and_list, __grains__),
        "opts": functools.partial(salt.utils.data.traverse_dict_and_list, __opts__),
    }
    for matcher, matchs in kwargs.items():
        t, matcher = matcher.split(":", 1)
        if "__env__" in matchs:
            matchs = {
                (pillarenv if k == "__env__" else k): v for k, v in matchs.items()
            }

        if t not in traverse:
            raise Exception(
                'Unknown traverse option "{}", should be one of {}'.format(
                    t, traverse.keys()
                )
            )
        cfgs = matchs.get(traverse[t](matcher, None), [])
        if not isinstance(cfgs, list):
            cfgs = [cfgs]
        stack_config_files += [cfg.replace("__env__", pillarenv) for cfg in cfgs]

    for cfg in stack_config_files:
        if not os.path.isfile(cfg):
            log.info('Ignoring pillar stack cfg "%s": file does not exist', cfg)
            continue
        stack = _process_stack_cfg(cfg, stack, minion_id, pillar)
    return stack


def _to_unix_slashes(path):
    return posixpath.join(*path.split(os.sep))


def _process_stack_cfg(cfg, stack, minion_id, pillar):
    log.debug("Pillar.stack config: %s", cfg)
    basedir, filename = os.path.split(cfg)
    jenv = Environment(
        loader=FileSystemLoader(basedir),
        extensions=["jinja2.ext.do", salt.utils.jinja.SerializerExtension],
    )
    jenv.globals.update(
        {
            "__opts__": __opts__,
            "__salt__": __salt__,
            "__grains__": __grains__,
            "__stack__": {
                "traverse": salt.utils.data.traverse_dict_and_list,
                "cfg_path": cfg,
            },
            "minion_id": minion_id,
            "pillar": pillar,
            "pillarenv": __opts__["pillarenv"],
        }
    )
    for item in _parse_stack_cfg(jenv.get_template(filename).render(stack=stack)):
        if not item.strip():
            continue  # silently ignore whitespace or empty lines
        paths = glob.glob(os.path.join(basedir, item))
        if not paths:
            log.info(
                'Ignoring pillar stack template "%s": can\'t find from root dir "%s"',
                item,
                basedir,
            )
            continue
        for path in sorted(paths):
            log.debug("YAML: basedir=%s, path=%s", basedir, path)
            # FileSystemLoader always expects unix-style paths
            unix_path = _to_unix_slashes(os.path.relpath(path, basedir))
            try:
                yaml = jenv.get_template(unix_path).render(stack=stack, ymlpath=path)
            except Exception as e:
                raise Exception(f'Stack pillar template render error in {path}:\n"{e}"')
            try:
                obj = salt.utils.yaml.safe_load(yaml)
            except Exception as e:
                raise Exception(
                    "Stack pillar yaml parsing error in {}:\n{}\n{}".format(
                        path, e, yaml
                    )
                )
            if isinstance(obj, dict):
                stack = _merge_dict(stack, obj)
    return stack


def _cleanup(obj):
    if obj:
        if isinstance(obj, dict):
            obj.pop("__", None)
            for k, v in obj.items():
                obj[k] = _cleanup(v)
        elif isinstance(obj, list) and isinstance(obj[0], dict) and "__" in obj[0]:
            del obj[0]
    return obj


def _merge_dict(stack, obj):
    strategy = obj.pop("__", "merge-last")
    if strategy not in strategies:
        raise Exception(f'Unknown strategy "{strategy}", should be one of {strategies}')
    if strategy == "overwrite":
        return _cleanup(obj)
    else:
        for k, v in obj.items():
            if strategy == "remove":
                stack.pop(k, None)
                continue
            if k in stack:
                if strategy == "merge-first":
                    # merge-first is same as merge-last but the other way round
                    # so let's switch stack[k] and v
                    stack_k = stack[k]
                    stack[k] = _cleanup(v)
                    v = stack_k
                if type(stack[k]) != type(v):
                    log.debug(
                        "Force overwrite, types differ: '%s' != '%s'", stack[k], v
                    )
                    stack[k] = _cleanup(v)
                elif isinstance(v, dict):
                    stack[k] = _merge_dict(stack[k], v)
                elif isinstance(v, list):
                    stack[k] = _merge_list(stack[k], v)
                else:
                    stack[k] = v
            else:
                stack[k] = _cleanup(v)
        return stack


def _merge_list(stack, obj):
    strategy = "merge-last"
    if obj and isinstance(obj[0], dict) and "__" in obj[0]:
        strategy = obj[0]["__"]
        del obj[0]
    if strategy not in strategies:
        raise Exception(f'Unknown strategy "{strategy}", should be one of {strategies}')
    if strategy == "overwrite":
        return obj
    elif strategy == "remove":
        return [item for item in stack if item not in obj]
    elif strategy == "merge-first":
        return obj + stack
    else:
        return stack + obj


def _parse_stack_cfg(content):
    """
    Allow top level cfg to be YAML
    """
    try:
        obj = salt.utils.yaml.safe_load(content)
        if isinstance(obj, list):
            return obj
    except Exception as e:  # pylint: disable=broad-except
        pass
    return content.splitlines()



import pytest

import salt.pillar.consul_pillar as consul_pillar
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skipif(
        not consul_pillar.consul, reason="python-consul module not installed"
    )
]


@pytest.fixture
def base_pillar_data():
    return [
        {
            "Value": "/path/to/certs/testsite1.crt",
            "Key": "test-shared/sites/testsite1/ssl/certs/SSLCertificateFile",
        },
        {
            "Value": "/path/to/certs/testsite1.key",
            "Key": "test-shared/sites/testsite1/ssl/certs/SSLCertificateKeyFile",
        },
        {"Value": None, "Key": "test-shared/sites/testsite1/ssl/certs/"},
        {"Value": "True", "Key": "test-shared/sites/testsite1/ssl/force"},
        {"Value": None, "Key": "test-shared/sites/testsite1/ssl/"},
        {
            "Value": "salt://sites/testsite1.tmpl",
            "Key": "test-shared/sites/testsite1/template",
        },
        {"Value": "test.example.com", "Key": "test-shared/sites/testsite1/uri"},
        {"Value": None, "Key": "test-shared/sites/testsite1/"},
        {"Value": None, "Key": "test-shared/sites/"},
        {"Value": "Test User", "Key": "test-shared/user/full_name"},
        {"Value": "adm\nwww-data\nmlocate", "Key": "test-shared/user/groups"},
        {"Value": '"adm\nwww-data\nmlocate"', "Key": "test-shared/user/dontsplit"},
        {"Value": "yaml:\n  key: value\n", "Key": "test-shared/user/dontexpand"},
        {"Value": None, "Key": "test-shared/user/blankvalue"},
        {"Value": "test", "Key": "test-shared/user/login"},
        {"Value": None, "Key": "test-shared/user/"},
    ]


@pytest.fixture
def configure_loader_modules():
    return {
        consul_pillar: {
            "__opts__": {
                "consul_config": {"consul.port": 8500, "consul.host": "172.17.0.15"}
            },
            "get_conn": MagicMock(return_value="consul_connection"),
        }
    }


def test_connection(base_pillar_data):
    with patch.dict(consul_pillar.__salt__, {"grains.get": MagicMock(return_value={})}):
        with patch.object(
            consul_pillar,
            "consul_fetch",
            MagicMock(return_value=("2232", base_pillar_data)),
        ):
            consul_pillar.ext_pillar(
                "testminion", {}, "consul_config root=test-shared/"
            )
            consul_pillar.get_conn.assert_called_once_with(
                consul_pillar.__opts__, "consul_config"
            )


def test_pillar_data(base_pillar_data):
    with patch.dict(consul_pillar.__salt__, {"grains.get": MagicMock(return_value={})}):
        with patch.object(
            consul_pillar,
            "consul_fetch",
            MagicMock(return_value=("2232", base_pillar_data)),
        ):
            pillar_data = consul_pillar.ext_pillar(
                "testminion", {}, "consul_config root=test-shared/"
            )
            consul_pillar.consul_fetch.assert_called_once_with(
                "consul_connection", "test-shared"
            )
            assert sorted(pillar_data) == ["sites", "user"]
            assert "blankvalue" not in pillar_data["user"]


def test_blank_root(base_pillar_data):
    with patch.dict(consul_pillar.__salt__, {"grains.get": MagicMock(return_value={})}):
        with patch.object(
            consul_pillar,
            "consul_fetch",
            MagicMock(return_value=("2232", base_pillar_data)),
        ):
            pillar_data = consul_pillar.ext_pillar("testminion", {}, "consul_config")
            consul_pillar.consul_fetch.assert_called_once_with("consul_connection", "")
            assert sorted(pillar_data) == ["test-shared"]


def test_pillar_nest(base_pillar_data):
    with patch.dict(consul_pillar.__salt__, {"grains.get": MagicMock(return_value={})}):
        with patch.object(
            consul_pillar,
            "consul_fetch",
            MagicMock(return_value=("2232", base_pillar_data)),
        ):
            pillar_data = consul_pillar.ext_pillar(
                "testminion",
                {},
                "consul_config pillar_root=nested-key/ root=test-shared/ ",
            )
            assert sorted(pillar_data["nested-key"]) == ["sites", "user"]
            assert "blankvalue" not in pillar_data["nested-key"]["user"]


def test_value_parsing(base_pillar_data):
    with patch.dict(consul_pillar.__salt__, {"grains.get": MagicMock(return_value={})}):
        with patch.object(
            consul_pillar,
            "consul_fetch",
            MagicMock(return_value=("2232", base_pillar_data)),
        ):
            pillar_data = consul_pillar.ext_pillar(
                "testminion", {}, "consul_config root=test-shared/"
            )
            assert isinstance(pillar_data["user"]["dontsplit"], str)


def test_non_expansion(base_pillar_data):
    with patch.dict(consul_pillar.__salt__, {"grains.get": MagicMock(return_value={})}):
        with patch.object(
            consul_pillar,
            "consul_fetch",
            MagicMock(return_value=("2232", base_pillar_data)),
        ):
            pillar_data = consul_pillar.ext_pillar(
                "testminion",
                {},
                "consul_config root=test-shared/ expand_keys=false",
            )
            assert isinstance(pillar_data["user"]["dontexpand"], str)


def test_dict_merge():
    test_dict = {}
    simple_dict = {"key1": {"key2": "val1"}}
    with patch.dict(test_dict, simple_dict):
        assert consul_pillar.dict_merge(test_dict, simple_dict) == simple_dict
    with patch.dict(test_dict, {"key1": {"key3": {"key4": "value"}}}):
        assert consul_pillar.dict_merge(test_dict, simple_dict) == {
            "key1": {"key2": "val1", "key3": {"key4": "value"}}
        }



"""
Module for working with the Glassfish/Payara 4.x management API
.. versionadded:: 2016.11.0
:depends: requests
"""

import urllib.parse

import salt.defaults.exitcodes
import salt.utils.json
from salt.exceptions import CommandExecutionError

try:
    import requests

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False

__virtualname__ = "glassfish"

# Default server
DEFAULT_SERVER = {
    "ssl": False,
    "url": "localhost",
    "port": 4848,
    "user": None,
    "password": None,
}


def __virtual__():
    """
    Only load if requests is installed
    """
    if HAS_LIBS:
        return __virtualname__
    else:
        return (
            False,
            'The "{}" module could not be loaded: "requests" is not installed.'.format(
                __virtualname__
            ),
        )


def _get_headers():
    """
    Return fixed dict with headers (JSON data + mandatory "Requested by" header)
    """
    return {
        "Accept": "application/json",
        "Content-Type": "application/json",
        "X-Requested-By": "GlassFish REST HTML interface",
    }


def _get_auth(username, password):
    """
    Returns the HTTP auth header
    """
    if username and password:
        return requests.auth.HTTPBasicAuth(username, password)
    else:
        return None


def _get_url(ssl, url, port, path):
    """
    Returns the URL of the endpoint
    """
    if ssl:
        return f"https://{url}:{port}/management/domain/{path}"
    else:
        return f"http://{url}:{port}/management/domain/{path}"


def _get_server(server):
    """
    Returns the server information if provided, or the defaults
    """
    return server if server else DEFAULT_SERVER


def _clean_data(data):
    """
    Removes SaltStack params from **kwargs
    """
    for key in list(data):
        if key.startswith("__pub"):
            del data[key]
    return data


def _api_response(response):
    """
    Check response status code + success_code returned by glassfish
    """
    if response.status_code == 404:
        __context__["retcode"] = salt.defaults.exitcodes.SALT_BUILD_FAIL
        raise CommandExecutionError("Element doesn't exists")
    if response.status_code == 401:
        __context__["retcode"] = salt.defaults.exitcodes.SALT_BUILD_FAIL
        raise CommandExecutionError("Bad username or password")
    elif response.status_code == 200 or response.status_code == 500:
        try:
            data = salt.utils.json.loads(response.content)
            if data["exit_code"] != "SUCCESS":
                __context__["retcode"] = salt.defaults.exitcodes.SALT_BUILD_FAIL
                raise CommandExecutionError(data["message"])
            return data
        except ValueError:
            __context__["retcode"] = salt.defaults.exitcodes.SALT_BUILD_FAIL
            raise CommandExecutionError("The server returned no data")
    else:
        response.raise_for_status()


def _api_get(path, server=None):
    """
    Do a GET request to the API
    """
    server = _get_server(server)
    response = requests.get(
        url=_get_url(server["ssl"], server["url"], server["port"], path),
        auth=_get_auth(server["user"], server["password"]),
        headers=_get_headers(),
        verify=True,
        timeout=120,
    )
    return _api_response(response)


def _api_post(path, data, server=None):
    """
    Do a POST request to the API
    """
    server = _get_server(server)
    response = requests.post(
        url=_get_url(server["ssl"], server["url"], server["port"], path),
        auth=_get_auth(server["user"], server["password"]),
        headers=_get_headers(),
        data=salt.utils.json.dumps(data),
        verify=True,
        timeout=120,
    )
    return _api_response(response)


def _api_delete(path, data, server=None):
    """
    Do a DELETE request to the API
    """
    server = _get_server(server)
    response = requests.delete(
        url=_get_url(server["ssl"], server["url"], server["port"], path),
        auth=_get_auth(server["user"], server["password"]),
        headers=_get_headers(),
        params=data,
        verify=True,
        timeout=120,
    )
    return _api_response(response)


# "Middle layer": uses _api_* functions to enum/get/create/update/delete elements
def _enum_elements(name, server=None):
    """
    Enum elements
    """
    elements = []
    data = _api_get(name, server)

    if any(data["extraProperties"]["childResources"]):
        for element in data["extraProperties"]["childResources"]:
            elements.append(element)
        return elements
    return None


def _get_element_properties(name, element_type, server=None):
    """
    Get an element's properties
    """
    properties = {}
    data = _api_get(f"{element_type}/{name}/property", server)

    # Get properties into a dict
    if any(data["extraProperties"]["properties"]):
        for element in data["extraProperties"]["properties"]:
            properties[element["name"]] = element["value"]
        return properties
    return {}


def _get_element(name, element_type, server=None, with_properties=True):
    """
    Get an element with or without properties
    """
    element = {}
    name = urllib.parse.quote(name, safe="")
    data = _api_get(f"{element_type}/{name}", server)

    # Format data, get properties if asked, and return the whole thing
    if any(data["extraProperties"]["entity"]):
        for key, value in data["extraProperties"]["entity"].items():
            element[key] = value
        if with_properties:
            element["properties"] = _get_element_properties(name, element_type)
        return element
    return None


def _create_element(name, element_type, data, server=None):
    """
    Create a new element
    """
    # Define property and id from name and properties + remove SaltStack parameters
    if "properties" in data:
        data["property"] = ""
        for key, value in data["properties"].items():
            if not data["property"]:
                data["property"] += "{}={}".format(key, value.replace(":", "\\:"))
            else:
                data["property"] += ":{}={}".format(key, value.replace(":", "\\:"))
        del data["properties"]

    # Send request
    _api_post(element_type, _clean_data(data), server)
    return urllib.parse.unquote(name)


def _update_element(name, element_type, data, server=None):
    """
    Update an element, including its properties
    """
    # Urlencode the name (names may have slashes)
    name = urllib.parse.quote(name, safe="")

    # Update properties first
    if "properties" in data:
        properties = []
        for key, value in data["properties"].items():
            properties.append({"name": key, "value": value})
        _api_post(f"{element_type}/{name}/property", properties, server)
        del data["properties"]

        # If the element only contained properties
        if not data:
            return urllib.parse.unquote(name)

    # Get the current data then merge updated data into it
    update_data = _get_element(name, element_type, server, with_properties=False)
    if update_data:
        update_data.update(data)
    else:
        __context__["retcode"] = salt.defaults.exitcodes.SALT_BUILD_FAIL
        raise CommandExecutionError(f"Cannot update {name}")

    # Finally, update the element
    _api_post(f"{element_type}/{name}", _clean_data(update_data), server)
    return urllib.parse.unquote(name)


def _delete_element(name, element_type, data, server=None):
    """
    Delete an element
    """
    _api_delete(
        "{}/{}".format(element_type, urllib.parse.quote(name, safe="")), data, server
    )
    return name


# Connector connection pools
def enum_connector_c_pool(server=None):
    """
    Enum connection pools
    """
    return _enum_elements("resources/connector-connection-pool", server)


def get_connector_c_pool(name, server=None):
    """
    Get a specific connection pool
    """
    return _get_element(name, "resources/connector-connection-pool", server)


def create_connector_c_pool(name, server=None, **kwargs):
    """
    Create a connection pool
    """
    defaults = {
        "connectionDefinitionName": "javax.jms.ConnectionFactory",
        "resourceAdapterName": "jmsra",
        "associateWithThread": False,
        "connectionCreationRetryAttempts": 0,
        "connectionCreationRetryIntervalInSeconds": 0,
        "connectionLeakReclaim": False,
        "connectionLeakTimeoutInSeconds": 0,
        "description": "",
        "failAllConnections": False,
        "id": name,
        "idleTimeoutInSeconds": 300,
        "isConnectionValidationRequired": False,
        "lazyConnectionAssociation": False,
        "lazyConnectionEnlistment": False,
        "matchConnections": True,
        "maxConnectionUsageCount": 0,
        "maxPoolSize": 32,
        "maxWaitTimeInMillis": 60000,
        "ping": False,
        "poolResizeQuantity": 2,
        "pooling": True,
        "steadyPoolSize": 8,
        "target": "server",
        "transactionSupport": "",
        "validateAtmostOncePeriodInSeconds": 0,
    }

    # Data = defaults + merge kwargs + remove salt
    data = defaults
    data.update(kwargs)

    # Check TransactionSupport against acceptable values
    if data["transactionSupport"] and data["transactionSupport"] not in (
        "XATransaction",
        "LocalTransaction",
        "NoTransaction",
    ):
        raise CommandExecutionError("Invalid transaction support")

    return _create_element(name, "resources/connector-connection-pool", data, server)


def update_connector_c_pool(name, server=None, **kwargs):
    """
    Update a connection pool
    """
    if "transactionSupport" in kwargs and kwargs["transactionSupport"] not in (
        "XATransaction",
        "LocalTransaction",
        "NoTransaction",
    ):
        raise CommandExecutionError("Invalid transaction support")
    return _update_element(name, "resources/connector-connection-pool", kwargs, server)


def delete_connector_c_pool(name, target="server", cascade=True, server=None):
    """
    Delete a connection pool
    """
    data = {"target": target, "cascade": cascade}
    return _delete_element(name, "resources/connector-connection-pool", data, server)


# Connector resources
def enum_connector_resource(server=None):
    """
    Enum connection resources
    """
    return _enum_elements("resources/connector-resource", server)


def get_connector_resource(name, server=None):
    """
    Get a specific connection resource
    """
    return _get_element(name, "resources/connector-resource", server)


def create_connector_resource(name, server=None, **kwargs):
    """
    Create a connection resource
    """
    defaults = {
        "description": "",
        "enabled": True,
        "id": name,
        "poolName": "",
        "objectType": "user",
        "target": "server",
    }

    # Data = defaults + merge kwargs + poolname
    data = defaults
    data.update(kwargs)

    if not data["poolName"]:
        raise CommandExecutionError("No pool name!")

    # Fix for lowercase vs camelCase naming differences
    for key, value in list(data.items()):
        del data[key]
        data[key.lower()] = value

    return _create_element(name, "resources/connector-resource", data, server)


def update_connector_resource(name, server=None, **kwargs):
    """
    Update a connection resource
    """
    # You're not supposed to update jndiName, if you do so, it will crash, silently
    if "jndiName" in kwargs:
        del kwargs["jndiName"]
    return _update_element(name, "resources/connector-resource", kwargs, server)


def delete_connector_resource(name, target="server", server=None):
    """
    Delete a connection resource
    """
    return _delete_element(
        name, "resources/connector-resource", {"target": target}, server
    )


# JMS Destinations
def enum_admin_object_resource(server=None):
    """
    Enum JMS destinations
    """
    return _enum_elements("resources/admin-object-resource", server)


def get_admin_object_resource(name, server=None):
    """
    Get a specific JMS destination
    """
    return _get_element(name, "resources/admin-object-resource", server)


def create_admin_object_resource(name, server=None, **kwargs):
    """
    Create a JMS destination
    """
    defaults = {
        "description": "",
        "className": "com.sun.messaging.Queue",
        "enabled": True,
        "id": name,
        "resAdapter": "jmsra",
        "resType": "javax.jms.Queue",
        "target": "server",
    }

    # Data = defaults + merge kwargs + poolname
    data = defaults
    data.update(kwargs)

    # ClassName isn't optional, even if the API says so
    if data["resType"] == "javax.jms.Queue":
        data["className"] = "com.sun.messaging.Queue"
    elif data["resType"] == "javax.jms.Topic":
        data["className"] = "com.sun.messaging.Topic"
    else:
        raise CommandExecutionError(
            'resType should be "javax.jms.Queue" or "javax.jms.Topic"!'
        )

    if data["resAdapter"] != "jmsra":
        raise CommandExecutionError('resAdapter should be "jmsra"!')

    # Fix for lowercase vs camelCase naming differences
    if "resType" in data:
        data["restype"] = data["resType"]
        del data["resType"]
    if "className" in data:
        data["classname"] = data["className"]
        del data["className"]

    return _create_element(name, "resources/admin-object-resource", data, server)


def update_admin_object_resource(name, server=None, **kwargs):
    """
    Update a JMS destination
    """
    if "jndiName" in kwargs:
        del kwargs["jndiName"]
    return _update_element(name, "resources/admin-object-resource", kwargs, server)


def delete_admin_object_resource(name, target="server", server=None):
    """
    Delete a JMS destination
    """
    return _delete_element(
        name, "resources/admin-object-resource", {"target": target}, server
    )


# JDBC Pools
def enum_jdbc_connection_pool(server=None):
    """
    Enum JDBC pools
    """
    return _enum_elements("resources/jdbc-connection-pool", server)


def get_jdbc_connection_pool(name, server=None):
    """
    Get a specific JDBC pool
    """
    return _get_element(name, "resources/jdbc-connection-pool", server)


def create_jdbc_connection_pool(name, server=None, **kwargs):
    """
    Create a connection resource
    """
    defaults = {
        "allowNonComponentCallers": False,
        "associateWithThread": False,
        "connectionCreationRetryAttempts": "0",
        "connectionCreationRetryIntervalInSeconds": "10",
        "connectionLeakReclaim": False,
        "connectionLeakTimeoutInSeconds": "0",
        "connectionValidationMethod": "table",
        "datasourceClassname": "",
        "description": "",
        "driverClassname": "",
        "failAllConnections": False,
        "idleTimeoutInSeconds": "300",
        "initSql": "",
        "isConnectionValidationRequired": False,
        "isIsolationLevelGuaranteed": True,
        "lazyConnectionAssociation": False,
        "lazyConnectionEnlistment": False,
        "matchConnections": False,
        "maxConnectionUsageCount": "0",
        "maxPoolSize": "32",
        "maxWaitTimeInMillis": 60000,
        "name": name,
        "nonTransactionalConnections": False,
        "ping": False,
        "poolResizeQuantity": "2",
        "pooling": True,
        "resType": "",
        "sqlTraceListeners": "",
        "statementCacheSize": "0",
        "statementLeakReclaim": False,
        "statementLeakTimeoutInSeconds": "0",
        "statementTimeoutInSeconds": "-1",
        "steadyPoolSize": "8",
        "target": "server",
        "transactionIsolationLevel": "",
        "validateAtmostOncePeriodInSeconds": "0",
        "validationClassname": "",
        "validationTableName": "",
        "wrapJdbcObjects": True,
    }

    # Data = defaults + merge kwargs + poolname
    data = defaults
    data.update(kwargs)

    # Check resType against acceptable values
    if data["resType"] not in (
        "javax.sql.DataSource",
        "javax.sql.XADataSource",
        "javax.sql.ConnectionPoolDataSource",
        "java.sql.Driver",
    ):
        raise CommandExecutionError("Invalid resource type")

    # Check connectionValidationMethod against acceptable velues
    if data["connectionValidationMethod"] not in (
        "auto-commit",
        "meta-data",
        "table",
        "custom-validation",
    ):
        raise CommandExecutionError("Invalid connection validation method")

    if data["transactionIsolationLevel"] and data["transactionIsolationLevel"] not in (
        "read-uncommitted",
        "read-committed",
        "repeatable-read",
        "serializable",
    ):
        raise CommandExecutionError("Invalid transaction isolation level")

    if not data["datasourceClassname"] and data["resType"] in (
        "javax.sql.DataSource",
        "javax.sql.ConnectionPoolDataSource",
        "javax.sql.XADataSource",
    ):
        raise CommandExecutionError(
            "No datasource class name while using datasource resType"
        )
    if not data["driverClassname"] and data["resType"] == "java.sql.Driver":
        raise CommandExecutionError("No driver class nime while using driver resType")

    return _create_element(name, "resources/jdbc-connection-pool", data, server)


def update_jdbc_connection_pool(name, server=None, **kwargs):
    """
    Update a JDBC pool
    """
    return _update_element(name, "resources/jdbc-connection-pool", kwargs, server)


def delete_jdbc_connection_pool(name, target="server", cascade=False, server=None):
    """
    Delete a JDBC pool
    """
    data = {"target": target, "cascade": cascade}
    return _delete_element(name, "resources/jdbc-connection-pool", data, server)


# JDBC resources
def enum_jdbc_resource(server=None):
    """
    Enum JDBC resources
    """
    return _enum_elements("resources/jdbc-resource", server)


def get_jdbc_resource(name, server=None):
    """
    Get a specific JDBC resource
    """
    return _get_element(name, "resources/jdbc-resource", server)


def create_jdbc_resource(name, server=None, **kwargs):
    """
    Create a JDBC resource
    """
    defaults = {
        "description": "",
        "enabled": True,
        "id": name,
        "poolName": "",
        "target": "server",
    }

    # Data = defaults + merge kwargs + poolname
    data = defaults
    data.update(kwargs)

    if not data["poolName"]:
        raise CommandExecutionError("No pool name!")

    return _create_element(name, "resources/jdbc-resource", data, server)


def update_jdbc_resource(name, server=None, **kwargs):
    """
    Update a JDBC resource
    """
    # You're not supposed to update jndiName, if you do so, it will crash, silently
    if "jndiName" in kwargs:
        del kwargs["jndiName"]
    return _update_element(name, "resources/jdbc-resource", kwargs, server)


def delete_jdbc_resource(name, target="server", server=None):
    """
    Delete a JDBC resource
    """
    return _delete_element(name, "resources/jdbc-resource", {"target": target}, server)


# System properties
def get_system_properties(server=None):
    """
    Get system properties
    """
    properties = {}
    data = _api_get("system-properties", server)

    # Get properties into a dict
    if any(data["extraProperties"]["systemProperties"]):
        for element in data["extraProperties"]["systemProperties"]:
            properties[element["name"]] = element["value"]
        return properties
    return {}


def update_system_properties(data, server=None):
    """
    Update system properties
    """
    _api_post("system-properties", _clean_data(data), server)
    return data


def delete_system_properties(name, server=None):
    """
    Delete a system property
    """
    _api_delete(f"system-properties/{name}", None, server)



"""
    :codeauthor: Mike Place (mp@saltstack.com)


    tests.unit.modules.mysql
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import logging

import pytest

import salt.modules.mysql as mysql
from tests.support.mock import MagicMock, call, mock_open, patch

try:
    import pymysql

    HAS_PYMYSQL = True
except ImportError:
    HAS_PYMYSQL = False


log = logging.getLogger(__name__)

__all_privileges__ = [
    "ALTER",
    "ALTER ROUTINE",
    "BACKUP_ADMIN",
    "BINLOG_ADMIN",
    "CONNECTION_ADMIN",
    "CREATE",
    "CREATE ROLE",
    "CREATE ROUTINE",
    "CREATE TABLESPACE",
    "CREATE TEMPORARY TABLES",
    "CREATE USER",
    "CREATE VIEW",
    "DELETE",
    "DROP",
    "DROP ROLE",
    "ENCRYPTION_KEY_ADMIN",
    "EVENT",
    "EXECUTE",
    "FILE",
    "GROUP_REPLICATION_ADMIN",
    "INDEX",
    "INSERT",
    "LOCK TABLES",
    "PERSIST_RO_VARIABLES_ADMIN",
    "PROCESS",
    "REFERENCES",
    "RELOAD",
    "REPLICATION CLIENT",
    "REPLICATION SLAVE",
    "REPLICATION_SLAVE_ADMIN",
    "RESOURCE_GROUP_ADMIN",
    "RESOURCE_GROUP_USER",
    "ROLE_ADMIN",
    "SELECT",
    "SET_USER_ID",
    "SHOW DATABASES",
    "SHOW VIEW",
    "SHUTDOWN",
    "SUPER",
    "SYSTEM_VARIABLES_ADMIN",
    "TRIGGER",
    "UPDATE",
    "XA_RECOVER_ADMIN",
]

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skipif(
        mysql.MySQLdb is None, reason="No python mysql client installed."
    ),
]


class MockMySQLConnect:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def autocommit(self, *args, **kwards):
        return True


@pytest.fixture
def configure_loader_modules():
    return {mysql: {}}


def test_user_exists():
    """
    Test to see if mysql module properly forms the MySQL query to see if a user exists

    Do it before test_user_create_when_user_exists mocks the user_exists call
    """
    with patch.object(mysql, "version", return_value="8.0.10"):
        _test_call(
            mysql.user_exists,
            {
                "sql": (
                    "SELECT User,Host FROM mysql.user WHERE "
                    "User = %(user)s AND Host = %(host)s AND "
                    "Password = PASSWORD(%(password)s)"
                ),
                "sql_args": {
                    "host": "localhost",
                    "password": "BLUECOW",
                    "user": "mytestuser",
                },
            },
            user="mytestuser",
            host="localhost",
            password="BLUECOW",
        )

    with patch.object(mysql, "version", return_value="10.1.38-MariaDB"):
        _test_call(
            mysql.user_exists,
            {
                "sql": (
                    "SELECT User,Host FROM mysql.user WHERE "
                    "User = %(user)s AND Host = %(host)s AND "
                    "Password = PASSWORD(%(password)s)"
                ),
                "sql_args": {
                    "host": "localhost",
                    "password": "BLUECOW",
                    "user": "mytestuser",
                },
            },
            user="mytestuser",
            host="localhost",
            password="BLUECOW",
        )

    with patch.object(mysql, "version", return_value="8.0.11"):
        _test_call(
            mysql.user_exists,
            {
                "sql": (
                    "SELECT User,Host FROM mysql.user WHERE "
                    "User = %(user)s AND Host = %(host)s"
                ),
                "sql_args": {"host": "localhost", "user": "mytestuser"},
            },
            user="mytestuser",
            host="localhost",
            password="BLUECOW",
        )

    with patch.object(mysql, "version", return_value="8.0.11"):
        with patch.object(
            mysql,
            "__get_auth_plugin",
            MagicMock(return_value="mysql_native_password"),
        ):
            _test_call(
                mysql.user_exists,
                {
                    "sql": (
                        "SELECT User,Host FROM mysql.user WHERE "
                        "User = %(user)s AND Host = %(host)s AND "
                        "Password = %(password)s"
                    ),
                    "sql_args": {
                        "host": "%",
                        "password": "*1A01CF8FBE6425398935FB90359AD8B817399102",
                        "user": "mytestuser",
                    },
                },
                user="mytestuser",
                host="%",
                password="BLUECOW",
            )

    with patch.object(mysql, "version", return_value="10.2.21-MariaDB"):
        _test_call(
            mysql.user_exists,
            {
                "sql": (
                    "SELECT User,Host FROM mysql.user WHERE "
                    "User = %(user)s AND Host = %(host)s AND "
                    "Password = PASSWORD(%(password)s)"
                ),
                "sql_args": {
                    "host": "localhost",
                    "password": "BLUECOW",
                    "user": "mytestuser",
                },
            },
            user="mytestuser",
            host="localhost",
            password="BLUECOW",
        )

    with patch.object(
        mysql, "version", side_effect=["", "10.2.21-MariaDB", "10.2.21-MariaDB"]
    ):
        _test_call(
            mysql.user_exists,
            {
                "sql": (
                    "SELECT User,Host FROM mysql.user WHERE "
                    "User = %(user)s AND Host = %(host)s AND "
                    "Password = PASSWORD(%(password)s)"
                ),
                "sql_args": {
                    "host": "localhost",
                    "password": "new_pass",
                    "user": "root",
                },
            },
            user="root",
            host="localhost",
            password="new_pass",
            connection_user="root",
            connection_pass="old_pass",
        )

    # test_user_create_when_user_exists():
    # ensure we don't try to create a user when one already exists
    # mock the version of MySQL
    with patch.object(mysql, "version", return_value="8.0.10"):
        with patch.object(mysql, "user_exists", MagicMock(return_value=True)):
            with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
                ret = mysql.user_create("testuser")
                assert not ret

    # test_user_create_when_user_exists():
    # ensure we don't try to create a user when one already exists
    # mock the version of MySQL
    with patch.object(mysql, "version", return_value="8.0.11"):
        with patch.object(mysql, "user_exists", MagicMock(return_value=True)):
            with patch.object(mysql, "verify_login", MagicMock(return_value=True)):
                with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
                    ret = mysql.user_create("testuser")
                    assert not False


def test_user_create():
    """
    Test the creation of a MySQL user in mysql exec module
    """
    with patch.object(mysql, "version", return_value="8.0.10"):
        with patch.object(
            mysql,
            "__get_auth_plugin",
            MagicMock(return_value="mysql_native_password"),
        ):
            _test_call(
                mysql.user_create,
                {
                    "sql": "CREATE USER %(user)s@%(host)s IDENTIFIED BY %(password)s",
                    "sql_args": {
                        "password": "BLUECOW",
                        "user": "testuser",
                        "host": "localhost",
                    },
                },
                "testuser",
                password="BLUECOW",
            )

    with patch.object(mysql, "version", return_value="8.0.11"):
        with patch.object(
            mysql,
            "__get_auth_plugin",
            MagicMock(return_value="mysql_native_password"),
        ):
            _test_call(
                mysql.user_create,
                {
                    "sql": "CREATE USER %(user)s@%(host)s IDENTIFIED WITH %(auth_plugin)s BY %(password)s",
                    "sql_args": {
                        "password": "BLUECOW",
                        "auth_plugin": "mysql_native_password",
                        "user": "testuser",
                        "host": "localhost",
                    },
                },
                "testuser",
                password="BLUECOW",
            )

    # Test creating a user with passwordless=True and unix_socket=True
    with patch.object(mysql, "version", return_value="8.0.10"):
        with patch.object(mysql, "plugin_status", MagicMock(return_value="ACTIVE")):
            _test_call(
                mysql.user_create,
                {
                    "sql": "CREATE USER %(user)s@%(host)s IDENTIFIED WITH auth_socket",
                    "sql_args": {"user": "testuser", "host": "localhost"},
                },
                "testuser",
                allow_passwordless=True,
                unix_socket=True,
            )

    with patch.object(mysql, "version", return_value="10.2.21-MariaDB"):
        with patch.object(mysql, "plugin_status", MagicMock(return_value="ACTIVE")):
            _test_call(
                mysql.user_create,
                {
                    "sql": "CREATE USER %(user)s@%(host)s IDENTIFIED VIA unix_socket",
                    "sql_args": {"user": "testuser", "host": "localhost"},
                },
                "testuser",
                allow_passwordless=True,
                unix_socket=True,
            )

    with patch.object(mysql, "version", side_effect=["", "8.0.10", "8.0.10"]):
        with patch.object(
            mysql, "user_exists", MagicMock(return_value=False)
        ), patch.object(
            mysql,
            "__get_auth_plugin",
            MagicMock(return_value="mysql_native_password"),
        ):
            _test_call(
                mysql.user_create,
                {
                    "sql": "CREATE USER %(user)s@%(host)s IDENTIFIED BY %(password)s",
                    "sql_args": {
                        "password": "new_pass",
                        "user": "root",
                        "host": "localhost",
                    },
                },
                "root",
                password="new_pass",
                connection_user="root",
                connection_pass="old_pass",
            )


def test_user_chpass():
    """
    Test changing a MySQL user password in mysql exec module
    """
    connect_mock = MagicMock()
    with patch.object(mysql, "_connect", connect_mock):
        with patch.object(mysql, "version", return_value="8.0.10"):
            with patch.object(mysql, "user_exists", MagicMock(return_value=True)):
                with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
                    mysql.user_chpass("testuser", password="BLUECOW")
                    calls = (
                        call()
                        .cursor()
                        .execute(
                            "UPDATE mysql.user SET Password=PASSWORD(%(password)s) WHERE User=%(user)s AND Host = %(host)s;",
                            {
                                "password": "BLUECOW",
                                "user": "testuser",
                                "host": "localhost",
                            },
                        ),
                        call().cursor().execute("FLUSH PRIVILEGES;"),
                    )
                    connect_mock.assert_has_calls(calls, any_order=True)

    connect_mock = MagicMock()
    with patch.object(mysql, "_connect", connect_mock):
        with patch.object(mysql, "version", return_value="8.0.11"):
            with patch.object(mysql, "user_exists", MagicMock(return_value=True)):
                with patch.object(
                    mysql,
                    "__get_auth_plugin",
                    MagicMock(return_value="mysql_native_password"),
                ):
                    with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
                        mysql.user_chpass("testuser", password="BLUECOW")
                        calls = (
                            call()
                            .cursor()
                            .execute(
                                "ALTER USER %(user)s@%(host)s IDENTIFIED WITH %(auth_plugin)s BY %(password)s;",
                                {
                                    "password": "BLUECOW",
                                    "user": "testuser",
                                    "host": "localhost",
                                    "auth_plugin": "mysql_native_password",
                                },
                            ),
                            call().cursor().execute("FLUSH PRIVILEGES;"),
                        )
                        connect_mock.assert_has_calls(calls, any_order=True)

    connect_mock = MagicMock()
    with patch.object(mysql, "_connect", connect_mock):
        with patch.object(mysql, "version", side_effect=["", "8.0.11", "8.0.11"]):
            with patch.object(mysql, "user_exists", MagicMock(return_value=True)):
                with patch.object(
                    mysql,
                    "__get_auth_plugin",
                    MagicMock(return_value="mysql_native_password"),
                ):
                    with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
                        mysql.user_chpass(
                            "root",
                            password="new_pass",
                            connection_user="root",
                            connection_pass="old_pass",
                        )
                        calls = (
                            call()
                            .cursor()
                            .execute(
                                "ALTER USER %(user)s@%(host)s IDENTIFIED WITH %(auth_plugin)s BY %(password)s;",
                                {
                                    "password": "new_pass",
                                    "user": "root",
                                    "host": "localhost",
                                    "auth_plugin": "mysql_native_password",
                                },
                            ),
                            call().cursor().execute("FLUSH PRIVILEGES;"),
                        )
                        connect_mock.assert_has_calls(calls, any_order=True)


def test_user_remove():
    """
    Test the removal of a MySQL user in mysql exec module
    """
    with patch.object(mysql, "user_exists", MagicMock(return_value=True)):
        _test_call(
            mysql.user_remove,
            {
                "sql": "DROP USER %(user)s@%(host)s",
                "sql_args": {"user": "testuser", "host": "localhost"},
            },
            "testuser",
        )


def test_db_check():
    """
    Test MySQL db check function in mysql exec module
    """
    _test_call(
        mysql.db_check,
        "CHECK TABLE `test``'\" db`.`my``'\" table`",
        "test`'\" db",
        "my`'\" table",
    )


def test_db_repair():
    """
    Test MySQL db repair function in mysql exec module
    """
    _test_call(
        mysql.db_repair,
        "REPAIR TABLE `test``'\" db`.`my``'\" table`",
        "test`'\" db",
        "my`'\" table",
    )


def test_db_optimize():
    """
    Test MySQL db optimize function in mysql exec module
    """
    _test_call(
        mysql.db_optimize,
        "OPTIMIZE TABLE `test``'\" db`.`my``'\" table`",
        "test`'\" db",
        "my`'\" table",
    )


def test_db_remove():
    """
    Test MySQL db remove function in mysql exec module
    """
    with patch.object(mysql, "db_exists", MagicMock(return_value=True)):
        _test_call(mysql.db_remove, "DROP DATABASE `test``'\" db`;", "test`'\" db")


def test_db_tables():
    """
    Test MySQL db_tables function in mysql exec module
    """
    with patch.object(mysql, "db_exists", MagicMock(return_value=True)):
        _test_call(mysql.db_tables, "SHOW TABLES IN `test``'\" db`", "test`'\" db")


def test_db_exists():
    """
    Test MySQL db_exists function in mysql exec module
    """
    _test_call(
        mysql.db_exists,
        {
            "sql": "SHOW DATABASES LIKE %(dbname)s;",
            "sql_args": {"dbname": r"""test%_`" db"""},
        },
        'test%_`" db',
    )


def test_db_create():
    """
    Test MySQL db_create function in mysql exec module
    """
    _test_call(
        mysql.db_create,
        "CREATE DATABASE IF NOT EXISTS `test``'\" db`;",
        "test`'\" db",
    )


def test_alter_db():
    """
    Test MySQL alter_db function in mysql exec module
    """
    mock_get_db = {
        "character_set": "utf8",
        "collate": "utf8_unicode_ci",
        "name": "my_test",
    }
    mock = MagicMock(return_value=mock_get_db)
    with patch.object(mysql, "db_get", return_value=mock) as mock_db_get:
        _test_call(
            mysql.alter_db,
            "ALTER DATABASE `my_test` CHARACTER SET utf8 COLLATE utf8_unicode_ci;",
            "my_test",
            "utf8",
            "utf8_unicode_ci",
        )


def test_user_list():
    """
    Test MySQL user_list function in mysql exec module
    """
    _test_call(mysql.user_list, "SELECT User,Host FROM mysql.user")


def test_user_info():
    """
    Test to see if the mysql execution module correctly forms the SQL for information on a MySQL user.
    """
    _test_call(
        mysql.user_info,
        {
            "sql": "SELECT * FROM mysql.user WHERE User = %(user)s AND Host = %(host)s",
            "sql_args": {"host": "localhost", "user": "mytestuser"},
        },
        "mytestuser",
    )


def test_user_grants():
    """
    Test to ensure the mysql user_grants function returns properly formed SQL for a basic query
    """
    with patch.object(mysql, "user_exists", MagicMock(return_value=True)):
        _test_call(
            mysql.user_grants,
            {
                "sql": "SHOW GRANTS FOR %(user)s@%(host)s",
                "sql_args": {"host": "localhost", "user": "testuser"},
            },
            "testuser",
        )


def test_grant_exists_true():
    """
    Test to ensure that we can find a grant that exists
    """
    mock_grants = [
        "GRANT USAGE ON *.* TO 'testuser'@'%'",
        "GRANT SELECT, INSERT, UPDATE ON `testdb`.`testtableone` TO 'testuser'@'%'",
        "GRANT SELECT(column1,column2) ON `testdb`.`testtableone` TO 'testuser'@'%'",
        "GRANT SELECT(column1,column2), INSERT(column1,column2) ON `testdb`.`testtableone` TO 'testuser'@'%'",
        "GRANT SELECT(column1,column2), UPDATE ON `testdb`.`testtableone` TO 'testuser'@'%'",
        "GRANT SELECT ON `testdb`.`testtabletwo` TO 'testuser'@'%'",
        "GRANT SELECT ON `testdb`.`testtablethree` TO 'testuser'@'%'",
    ]
    with patch.object(mysql, "version", return_value="5.6.41"):
        mock = MagicMock(return_value=mock_grants)
        with patch.object(
            mysql, "user_grants", return_value=mock_grants
        ) as mock_user_grants:
            ret = mysql.grant_exists(
                "SELECT, INSERT, UPDATE", "testdb.testtableone", "testuser", "%"
            )
            assert ret


def test_grant_exists_false():
    """
    Test to ensure that we don't find a grant that doesn't exist
    """
    mock_grants = [
        "GRANT USAGE ON *.* TO 'testuser'@'%'",
        "GRANT SELECT, INSERT, UPDATE ON `testdb`.`testtableone` TO 'testuser'@'%'",
        "GRANT SELECT(column1,column2) ON `testdb`.`testtableone` TO 'testuser'@'%'",
        "GRANT SELECT(column1,column2), UPDATE ON `testdb`.`testtableone` TO 'testuser'@'%'",
        "GRANT SELECT ON `testdb`.`testtablethree` TO 'testuser'@'%'",
    ]
    with patch.object(mysql, "version", return_value="5.6.41"):
        mock = MagicMock(return_value=mock_grants)
        with patch.object(
            mysql, "user_grants", return_value=mock_grants
        ) as mock_user_grants:
            ret = mysql.grant_exists("SELECT", "testdb.testtabletwo", "testuser", "%")
            assert not ret


def test_grant_exists_all():
    """
    Test to ensure that we can find a grant that exists
    """
    mock_grants = ["GRANT ALL PRIVILEGES ON testdb.testtableone TO `testuser`@`%`"]
    with patch.object(mysql, "version", return_value="8.0.10"):
        mock = MagicMock(return_value=mock_grants)
        with patch.object(
            mysql, "user_grants", return_value=mock_grants
        ) as mock_user_grants:
            ret = mysql.grant_exists("ALL", "testdb.testtableone", "testuser", "%")
            assert ret

    with patch.object(mysql, "version", return_value="8.0.10"):
        mock = MagicMock(return_value=mock_grants)
        with patch.object(
            mysql, "user_grants", return_value=mock_grants
        ) as mock_user_grants:
            ret = mysql.grant_exists(
                "all privileges", "testdb.testtableone", "testuser", "%"
            )
            assert ret

    mock_grants = ["GRANT ALL PRIVILEGES ON testdb.testtableone TO `testuser`@`%`"]
    with patch.object(mysql, "version", return_value="5.6.41"):
        mock = MagicMock(return_value=mock_grants)
        with patch.object(
            mysql, "user_grants", return_value=mock_grants
        ) as mock_user_grants:
            ret = mysql.grant_exists(
                "ALL PRIVILEGES", "testdb.testtableone", "testuser", "%"
            )
            assert ret

    mock_grants = [
        "GRANT SELECT, INSERT, UPDATE, DELETE, CREATE, DROP, RELOAD, SHUTDOWN, PROCESS, FILE, REFERENCES, INDEX, ALTER, SHOW DATABASES, SUPER, CREATE TEMPORARY TABLES, LOCK TABLES, EXECUTE, REPLICATION SLAVE, REPLICATION CLIENT, CREATE VIEW, SHOW VIEW, CREATE ROUTINE, ALTER ROUTINE, CREATE USER, EVENT, TRIGGER, CREATE TABLESPACE, CREATE ROLE, DROP ROLE ON *.* TO `testuser`@`%`",
        "GRANT BACKUP_ADMIN,BINLOG_ADMIN,CONNECTION_ADMIN,ENCRYPTION_KEY_ADMIN,GROUP_REPLICATION_ADMIN,PERSIST_RO_VARIABLES_ADMIN,REPLICATION_SLAVE_ADMIN,RESOURCE_GROUP_ADMIN,RESOURCE_GROUP_USER,ROLE_ADMIN,SET_USER_ID,SYSTEM_VARIABLES_ADMIN,XA_RECOVER_ADMIN ON *.* TO `testuser`@`%`",
    ]
    with patch.object(mysql, "version", return_value="8.0.10"):
        mock = MagicMock(return_value=mock_grants)
        with patch.object(
            mysql, "user_grants", return_value=mock_grants
        ) as mock_user_grants:
            ret = mysql.grant_exists("ALL", "*.*", "testuser", "%")
            assert ret

    with patch.object(mysql, "version", return_value="8.0.10"):
        mock = MagicMock(return_value=mock_grants)
        with patch.object(
            mysql, "user_grants", return_value=mock_grants
        ) as mock_user_grants:
            ret = mysql.grant_exists("all privileges", "*.*", "testuser", "%")
            assert ret


@pytest.mark.skipif(True, reason="TODO: Mock up user_grants()")
def test_grant_add():
    """
    Test grant_add function in mysql exec module
    """
    _test_call(
        mysql.grant_add,
        "",
        "SELECT,INSERT,UPDATE",
        "database.*",
        "frank",
        "localhost",
    )


@pytest.mark.skipif(True, reason="TODO: Mock up user_grants()")
def test_grant_revoke():
    """
    Test grant revoke in mysql exec module
    """
    _test_call(
        mysql.grant_revoke,
        "",
        "SELECT,INSERT,UPDATE",
        "database.*",
        "frank",
        "localhost",
    )


def test_processlist():
    """
    Test processlist function in mysql exec module
    """
    _test_call(mysql.processlist, "SHOW FULL PROCESSLIST")


def test_get_master_status():
    """
    Test get_master_status in the mysql execution module
    """
    _test_call(mysql.get_master_status, "SHOW MASTER STATUS")


def test_get_slave_status():
    """
    Test get_slave_status in the mysql execution module
    """
    _test_call(mysql.get_slave_status, "SHOW SLAVE STATUS")


def test_get_slave_status_bad_server():
    """
    Test get_slave_status in the mysql execution module, simulating a broken server
    """
    connect_mock = MagicMock(return_value=None)
    with patch.object(mysql, "_connect", connect_mock):
        with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
            rslt = mysql.get_slave_status()
            connect_mock.assert_has_calls([call()])
            assert rslt == []


@pytest.mark.skip(
    reason="MySQL module claims this function is not ready for production"
)
def test_free_slave():
    pass


def test_query():
    _test_call(mysql.query, "SELECT * FROM testdb", "testdb", "SELECT * FROM testdb")


@pytest.mark.skipif(not HAS_PYMYSQL, reason="Could not import pymysql")
def test_query_error():
    connect_mock = MagicMock()
    with patch.object(mysql, "_connect", connect_mock):
        with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
            # Use the OperationalError from the salt mysql module because that
            # exception can come from either MySQLdb or pymysql
            side_effect = mysql.OperationalError(9999, "Something Went Wrong")
            with patch.object(mysql, "_execute", MagicMock(side_effect=side_effect)):
                mysql.query("testdb", "SELECT * FROM testdb")
        assert "mysql.error" in mysql.__context__
        expected = "MySQL Error 9999: Something Went Wrong"
        assert mysql.__context__["mysql.error"] == expected


def test_plugin_add():
    """
    Test the adding/installing a MySQL / MariaDB plugin
    """
    with patch.object(mysql, "plugin_status", MagicMock(return_value="")):
        _test_call(
            mysql.plugin_add,
            'INSTALL PLUGIN auth_socket SONAME "auth_socket.so"',
            "auth_socket",
        )


def test_plugin_remove():
    """
    Test the removing/uninstalling a MySQL / MariaDB plugin
    """
    with patch.object(mysql, "plugin_status", MagicMock(return_value="ACTIVE")):
        _test_call(
            mysql.plugin_remove,
            "UNINSTALL PLUGIN auth_socket",
            "auth_socket",
        )


def test_plugin_status():
    """
    Test checking the status of a MySQL / MariaDB plugin
    """
    _test_call(
        mysql.plugin_status,
        {
            "sql": "SELECT PLUGIN_STATUS FROM INFORMATION_SCHEMA.PLUGINS WHERE PLUGIN_NAME = %(name)s",
            "sql_args": {"name": "auth_socket"},
        },
        "auth_socket",
    )


def test_sanitize_comment():
    """
    Test comment sanitization
    """
    input_data = """/*
    multiline
    comment
    */
    CREATE TABLE test_update (a VARCHAR(25)); # end of line comment
    # example comment
    insert into test_update values ("some #hash value");            -- ending comment
    insert into test_update values ("crazy -- not comment"); -- another ending comment
    -- another comment type
    """
    expected_response = """CREATE TABLE test_update (a VARCHAR(25));

insert into test_update values ("some #hash value");
insert into test_update values ("crazy -- not comment");

"""
    output = mysql._sanitize_comments(input_data)
    assert output == expected_response

    input_data = """-- --------------------------------------------------------
                    -- SQL Commands to set up the pmadb as described in the documentation.
                    --
                    -- This file is meant for use with MySQL 5 and above!
                    --
                    -- This script expects the user pma to already be existing. If we would put a
                    -- line here to create them too many users might just use this script and end
                    -- up with having the same password for the controluser.
                    --
                    -- This user "pma" must be defined in config.inc.php (controluser/controlpass)
                    --
                    -- Please don't forget to set up the tablenames in config.inc.php
                    --
                    -- --------------------------------------------------------
                    --
                    CREATE DATABASE IF NOT EXISTS `phpmyadmin`
                      DEFAULT CHARACTER SET utf8 COLLATE utf8_bin;
                    USE phpmyadmin;
    """

    expected_response = """CREATE DATABASE IF NOT EXISTS `phpmyadmin`
                      DEFAULT CHARACTER SET utf8 COLLATE utf8_bin;
                    USE phpmyadmin;"""

    output = mysql._sanitize_comments(input_data)
    assert output == expected_response


def _test_call(function, expected_sql, *args, **kwargs):
    connect_mock = MagicMock()
    with patch.object(mysql, "_connect", connect_mock):
        with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
            function(*args, **kwargs)
            if isinstance(expected_sql, dict):
                calls = (
                    call()
                    .cursor()
                    .execute("{}".format(expected_sql["sql"]), expected_sql["sql_args"])
                )
            else:
                calls = call().cursor().execute(f"{expected_sql}")
            connect_mock.assert_has_calls((calls,), True)


def test_file_query():
    """
    Test file_query
    """
    with patch.object(mysql, "HAS_SQLPARSE", False):
        ret = mysql.file_query("database", "filename")
        assert not ret

    file_data = """-- --------------------------------------------------------
                   -- SQL Commands to set up the pmadb as described in the documentation.
                   --
                   -- This file is meant for use with MySQL 5 and above!
                   --
                   -- This script expects the user pma to already be existing. If we would put a
                   -- line here to create them too many users might just use this script and end
                   -- up with having the same password for the controluser.
                   --
                   -- This user "pma" must be defined in config.inc.php (controluser/controlpass)
                   --
                   -- Please don't forget to set up the tablenames in config.inc.php
                   --
                   -- --------------------------------------------------------
                   --
                   USE phpmyadmin;

                   --
                   -- Table structure for table `pma__bookmark`
                   --

                   CREATE TABLE IF NOT EXISTS `pma__bookmark` (
                     `id` int(10) unsigned NOT NULL auto_increment,
                     `dbase` varchar(255) NOT NULL default '',
                     `user` varchar(255) NOT NULL default '',
                     `label` varchar(255) COLLATE utf8_general_ci NOT NULL default '',
                     `query` text NOT NULL,
                     PRIMARY KEY  (`id`)
                   )
                     COMMENT='Bookmarks'
                     DEFAULT CHARACTER SET utf8 COLLATE utf8_bin;
    """

    side_effect = [
        {"query time": {"human": "0.4ms", "raw": "0.00038"}, "rows affected": 0},
        {"query time": {"human": "8.9ms", "raw": "0.00893"}, "rows affected": 0},
    ]
    expected = {
        "query time": {"human": "8.9ms", "raw": "0.00893"},
        "rows affected": 0,
    }

    with patch("os.path.exists", MagicMock(return_value=True)):
        with patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
            with patch.object(mysql, "query", side_effect=side_effect):
                ret = mysql.file_query("database", "filename")
                assert ret, expected


@pytest.mark.skipif(not HAS_PYMYSQL, reason="Could not import pymysql")
def test__connect_pymysql_exception():
    """
    Test the _connect function in the MySQL module
    """
    with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
        with patch(
            "MySQLdb.connect",
            side_effect=pymysql.err.InternalError(
                1698, "Access denied for user 'root'@'localhost'"
            ),
        ):
            ret = mysql._connect()
            assert "mysql.error" in mysql.__context__
            assert (
                mysql.__context__["mysql.error"]
                == "MySQL Error 1698: Access denied for user 'root'@'localhost'"
            )


def test__connect_mysqldb_exception():
    """
    Test the _connect function in the MySQL module
    """
    with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
        with patch(
            "MySQLdb.connect",
            side_effect=mysql.OperationalError(
                1698, "Access denied for user 'root'@'localhost'"
            ),
        ):
            ret = mysql._connect()
            assert "mysql.error" in mysql.__context__
            assert (
                mysql.__context__["mysql.error"]
                == "MySQL Error 1698: Access denied for user 'root'@'localhost'"
            )


def test__connect_mysqldb():
    """
    Test the _connect function in the MySQL module
    """
    mysqldb_connect_mock = MagicMock(autospec=True, return_value=MockMySQLConnect())
    with patch.dict(mysql.__salt__, {"config.option": MagicMock()}):
        with patch("MySQLdb.connect", mysqldb_connect_mock):
            mysql._connect()
            assert "mysql.error" not in mysql.__context__



"""
This is the default glob matcher function.
"""

import fnmatch
import logging

log = logging.getLogger(__name__)


def match(tgt, opts=None, minion_id=None):
    """
    Returns true if the passed glob matches the id
    """
    if not opts:
        opts = __opts__
    if not minion_id:
        minion_id = opts.get("minion_id", opts["id"])
    if not isinstance(tgt, str):
        return False

    return fnmatch.fnmatch(minion_id, tgt)



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_mako_renderer(state, state_tree):
    """
    Test mako renderer when running state.sls
    """
    sls_contents = """
    #!mako|yaml
    %for a in [1,2,3]:
    echo ${a}:
        cmd.run
    %endfor
    """
    with pytest.helpers.temp_file("issue-55124.sls", sls_contents, state_tree):
        ret = state.sls("issue-55124")
        for state_return in ret:
            assert state_return.result is True
            assert "echo" in state_return.id



"""
unittests for table outputter
"""

import pytest

import salt.output.table_out as table_out
import salt.utils.stringutils


@pytest.fixture
def configure_loader_modules():
    return {table_out: {}}


# The test data should include unicode chars, and in Python 2 there should
# be an example both of an encoded str type and an actual unicode type.
# Since unicode_literals is imported, we will achieve the former using
# salt.utils.stringutils.to_str and the latter by simply using a string
# literal.
@pytest.fixture
def data():
    return [
        {
            "Food": salt.utils.stringutils.to_str("яйца, бекон, колбаса и спам"),
            "Price": 5.99,
        },
        {"Food": "спам, спам, спам, яйца и спам", "Price": 3.99},
    ]


def test_output(data):
    ret = table_out.output(data)
    assert ret == (
        "    -----------------------------------------\n"
        "    |              Food             | Price |\n"
        "    -----------------------------------------\n"
        "    |  яйца, бекон, колбаса и спам  |  5.99 |\n"
        "    -----------------------------------------\n"
        "    | спам, спам, спам, яйца и спам |  3.99 |\n"
        "    -----------------------------------------"
    )



import subprocess
import types

import pytest
from pytestskipmarkers.utils import platform
from saltfactories.utils.functional import MultiStateResult

pytestmark = [
    pytest.mark.skip_on_windows,
]


@pytest.fixture
def files(tmp_path):
    return types.SimpleNamespace(
        fpath_1=tmp_path / "fpath_1.txt",
        fpath_2=tmp_path / "fpath_2.txt",
        fpath_3=tmp_path / "fpath_3.txt",
    )


@pytest.fixture
def state_name(files, salt_master):
    name = "some-state"
    sls_contents = f"""
    create-fpath-1-file:
      file.managed:
        - name: {files.fpath_1}

    create-fpath-2-file:
      file.managed:
        - name: {files.fpath_2}

    create-fpath-3-file:
      file.managed:
        - name: {files.fpath_3}
    """
    assert files.fpath_1.exists() is False
    assert files.fpath_2.exists() is False
    assert files.fpath_3.exists() is False
    with salt_master.state_tree.base.temp_file(f"{name}.sls", sls_contents):
        if not platform.is_windows() and not platform.is_darwin():
            subprocess.run(
                [
                    "chown",
                    "-R",
                    "salt:salt",
                    str(salt_master.state_tree.base.write_path),
                ],
                check=False,
            )
        yield name


def test_salt_state_file(salt_cli, salt_minion, salt_master, state_name, files):
    """
    Test state file
    """
    assert files.fpath_1.exists() is False
    assert files.fpath_2.exists() is False
    assert files.fpath_3.exists() is False
    assert salt_master.is_running()

    ret = salt_cli.run("state.apply", state_name, minion_tgt=salt_minion.id)
    assert ret.returncode == 0
    assert ret.data
    if ret.stdout and "Minion did not return" in ret.stdout:
        pytest.skip("Skipping test, state took too long to apply")

    for state_return in MultiStateResult(ret.data):
        assert state_return.result is True

    assert files.fpath_1.exists() is True
    assert files.fpath_2.exists() is True
    assert files.fpath_3.exists() is True



"""
A salt module for modifying the audit policies on the machine

Though this module does not set group policy for auditing, it displays how all
auditing configuration is applied on the machine, either set directly or via
local or domain group policy.

.. versionadded:: 2018.3.4
.. versionadded:: 2019.2.1

This module allows you to view and modify the audit settings as they are applied
on the machine. The audit settings are broken down into nine categories:

- Account Logon
- Account Management
- Detailed Tracking
- DS Access
- Logon/Logoff
- Object Access
- Policy Change
- Privilege Use
- System

The ``get_settings`` function will return the subcategories for all nine of
the above categories in one dictionary along with their auditing status.

To modify a setting you only need to specify the subcategory name and the value
you wish to set. Valid settings are:

- No Auditing
- Success
- Failure
- Success and Failure

CLI Example:

.. code-block:: bash

    # Get current state of all audit settings
    salt '*' auditpol.get_settings

    # Get the current state of all audit settings in the "Account Logon"
    # category
    salt '*' auditpol.get_settings category="Account Logon"

    # Get current state of the "Credential Validation" setting
    salt '*' auditpol.get_setting name="Credential Validation"

    # Set the state of the "Credential Validation" setting to Success and
    # Failure
    salt '*' auditpol.set_setting name="Credential Validation" value="Success and Failure"

    # Set the state of the "Credential Validation" setting to No Auditing
    salt '*' auditpol.set_setting name="Credential Validation" value="No Auditing"
"""

import salt.utils.platform

# Define the module's virtual name
__virtualname__ = "auditpol"


def __virtual__():
    """
    Only works on Windows systems
    """
    if not salt.utils.platform.is_windows():
        return False, "Module win_auditpol: module only available on Windows"

    return __virtualname__


def get_settings(category="All"):
    """
    Get the current configuration for all audit settings specified in the
    category

    Args:
        category (:obj:`str`, optional):
            One of the nine categories to return. Can also be ``All`` to return
            the settings for all categories. Valid options are:

            - Account Logon
            - Account Management
            - Detailed Tracking
            - DS Access
            - Logon/Logoff
            - Object Access
            - Policy Change
            - Privilege Use
            - System
            - All

            Default is ``All``.

    Returns:
        dict: A dictionary containing all subcategories for the specified
            category along with their current configuration

    Raises:
        KeyError: On invalid category
        CommandExecutionError: If an error is encountered retrieving the settings

    CLI Example:

    .. code-block:: bash

        # Get current state of all audit settings
        salt '*' auditipol.get_settings

        # Get the current state of all audit settings in the "Account Logon"
        # category
        salt '*' auditpol.get_settings 'Account Logon'
    """
    return __utils__["auditpol.get_settings"](category=category)


def get_setting(name):
    """
    Get the current configuration for the named audit setting

    Args:
        name (str): The name of the setting to retrieve

    Returns:
        str: The current configuration for the named setting

    Raises:
        KeyError: On invalid setting name
        CommandExecutionError: If an error is encountered retrieving the settings

    CLI Example:

    .. code-block:: bash

        # Get current state of the "Credential Validation" setting
        salt '*' auditpol.get_setting 'Credential Validation'
    """
    return __utils__["auditpol.get_setting"](name=name)


def set_setting(name, value):
    """
    Set the configuration for the named audit setting

    Args:

        name (str):
            The name of the setting to configure

        value (str):
            The configuration for the named value. Valid options are:

            - No Auditing
            - Success
            - Failure
            - Success and Failure

    Returns:
        bool: ``True`` if successful

    Raises:
        KeyError: On invalid ``name`` or ``value``
        CommandExecutionError: If an error is encountered modifying the setting

    CLI Example:

    .. code-block:: bash

        # Set the state of the "Credential Validation" setting to Success and
        # Failure
        salt '*' auditpol.set_setting 'Credential Validation' 'Success and Failure'

        # Set the state of the "Credential Validation" setting to No Auditing
        salt '*' auditpol.set_setting 'Credential Validation' 'No Auditing'
    """
    return __utils__["auditpol.set_setting"](name=name, value=value)



r"""
Execution of Salt modules from within states
============================================

.. note::

    As of the 3005 release, you no longer need to opt-in to the new style of
    calling ``module.run``. The following config can be removed from ``/etc/salt/minion``:

    .. code-block:: yaml

        use_superseded:
          - module.run

    Both 'new' and 'legacy' styles of calling ``module.run`` are supported.


With `module.run` these states allow individual execution module calls to be
made via states. Here's a contrived example, to show you how it's done:

.. code-block:: yaml

    # New Style
    test.random_hash:
      module.run:
        - test.random_hash:
          - size: 42
          - hash_type: sha256

    # Legacy Style
    test.random_hash:
      module.run:
        - size: 42
        - hash_type: sha256

In the new style, the state ID (``test.random_hash``, in this case) is
irrelevant when using ``module.run``. It could have very well been written:

.. code-block:: yaml

    Generate a random hash:
      module.run:
        - test.random_hash:
          - size: 42
          - hash_type: sha256

For a simple state like that it's not a big deal, but if the module you're
using has certain parameters, things can get cluttered, fast. Using the
contrived custom module (stuck in ``/srv/salt/_modules/foo.py``, or your
configured file_roots_):

.. code-block:: python

    def bar(name, names, fun, state, saltenv):
        return "Name: {name} Names: {names} Fun: {fun} State: {state} Saltenv: {saltenv}".format(**locals())

Your legacy state has to look like this:

.. code-block:: yaml

    # Legacy style
    Unfortunate example:
      module.run:
      - name: foo.bar
      - m_name: Some name
      - m_names:
        - Such names
        - very wow
      - m_state: Arkansas
      - m_fun: Such fun
      - m_saltenv: Salty

With the new style it's much cleaner:

.. code-block:: yaml

    # New style
    Better:
      module.run:
      - foo.bar:
        - name: Some name
        - names:
          - Such names
          - very wow
        - state: Arkansas
        - fun: Such fun
        - saltenv: Salty

The new style also allows multiple modules in one state. For instance, you can
do this:

.. code-block:: yaml

    Do many things:
      module.run:
        - test.random_hash:
          - size: 10
          - hash_type: md5
        # Note the `:` at the end
        - test.true:
        - test.arg:
          - this
          - has
          - args
          - and: kwargs
          - isn't: that neat?
        # Note the `:` at the end, too
        - test.version:
        - test.fib:
          - 4

Where in the legacy style you would have had to split your states like this:

.. code-block:: yaml

    test.random_hash:
      module.run:
        - size: 10
        - hash_type: md5

    test.nop:
      module.run

    test.arg:
      module.run:
        - args:
          - this
          - has
          - args
        - kwargs:
            and: kwargs
            isn't: that neat?

    test.version:
      module.run

Another difference is that in the legacy style, unconsumed arguments to the
``module`` state were simply passed into the module function being executed:

.. code-block:: yaml

    show off module.run with args:
      module.run:
        - name: test.random_hash
        - size: 42
        - hash_type: sha256

The new style is much more explicit, with the arguments and keyword arguments
being nested under the name of the function:

.. code-block:: yaml

    show off module.run with args:
      module.run:
        # Note the lack of `name: `, and trailing `:`
        - test.random_hash:
          - size: 42
          - hash_type: sha256

If the function takes ``*args``, they can be passed in as well:

.. code-block:: yaml

    args and kwargs:
      module.run:
        - test.arg:
          - isn't
          - this
          - fun
          - this: that
          - salt: stack

Modern Examples
---------------

Here are some other examples using the modern ``module.run``:

.. code-block:: yaml

    fetch_out_of_band:
      module.run:
        - git.fetch:
          - cwd: /path/to/my/repo
          - user: myuser
          - opts: '--all'

A more complex example:

.. code-block:: yaml

    eventsviewer:
      module.run:
        - task.create_task:
          - name: events-viewer
          - user_name: System
          - action_type: Execute
          - cmd: 'c:\netops\scripts\events_viewer.bat'
          - trigger_type: 'Daily'
          - start_date: '2017-1-20'
          - start_time: '11:59PM'

It is sometimes desirable to trigger a function call after a state is executed,
for this the :mod:`module.wait <salt.states.module.wait>` state can be used:

.. code-block:: yaml

    add example to hosts:
      file.append:
        - name: /etc/hosts
        - text: 203.0.113.13     example.com

    # New Style
    mine.send:
      module.wait:
        # Again, note the trailing `:`
        - hosts.list_hosts:
        - watch:
          - file: add example to hosts

Legacy (Default) Examples
-------------------------

If you're using the legacy ``module.run``, due to how the state system works,
if a module function accepts an argument called, ``name``, then ``m_name`` must
be used to specify that argument, to avoid a collision with the ``name``
argument.

Here is a list of keywords hidden by the state system, which must be prefixed
with ``m_``:

* fun
* name
* names
* state
* saltenv

For example:

.. code-block:: yaml

    disable_nfs:
      module.run:
        - name: service.disable
        - m_name: nfs

Note that some modules read all or some of the arguments from a list of keyword
arguments. For example:

.. code-block:: yaml

    mine.send:
      module.run:
        - func: network.ip_addrs
        - kwargs:
            interface: eth0

.. code-block:: yaml

    cloud.create:
      module.run:
        - func: cloud.create
        - provider: test-provider
        - m_names:
          - test-vlad
        - kwargs: {
              ssh_username: 'ubuntu',
              image: 'ami-8d6d9daa',
              securitygroup: 'default',
              size: 'c3.large',
              location: 'ap-northeast-1',
              delvol_on_destroy: 'True'
          }

Other modules take the keyword arguments using this style:

.. code-block:: yaml

     mac_enable_ssh:
       module.run:
         - name: system.set_remote_login
         - enable: True

Another example that creates a recurring task that runs a batch file on a
Windows system:

.. code-block:: yaml

    eventsviewer:
      module.run:
        - name: task.create_task
        - m_name: 'events-viewer'
        - user_name: System
        - kwargs: {
              action_type: 'Execute',
              cmd: 'c:\netops\scripts\events_viewer.bat',
              trigger_type: 'Daily',
              start_date: '2017-1-20',
              start_time: '11:59PM'
        }

.. _file_roots: https://docs.saltproject.io/en/latest/ref/configuration/master.html#file-roots
"""

import logging

import salt.loader
import salt.utils.args
import salt.utils.functools
import salt.utils.jid
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


def wait(name, **kwargs):
    """
    Run a single module function only if the watch statement calls it

    ``name``
        The module function to execute

    ``**kwargs``
        Pass any arguments needed to execute the function

    .. note::
        Like the :mod:`cmd.run <salt.states.cmd.run>` state, this state will
        return ``True`` but not actually execute, unless one of the following
        two things happens:

        1. The state has a :ref:`watch requisite <requisites-watch>`, and
           the state which it is watching changes.

        2. Another state has a :ref:`watch_in requisite
           <requisites-watch-in>` which references this state, and the state
           wth the ``watch_in`` changes.
    """
    return {"name": name, "changes": {}, "result": True, "comment": ""}


# Alias module.watch to module.wait
watch = salt.utils.functools.alias_function(wait, "watch")


def run(**kwargs):
    """
    Run a single module function or a range of module functions in a batch.
    Supersedes ``module.run`` function, which requires ``m_`` prefix to
    function-specific parameters.

    :param returner:
        Specify a common returner for the whole batch to send the return data

    :param kwargs:
        Pass any arguments needed to execute the function(s)

    .. code-block:: yaml

      some_id_of_state:
        module.run:
          - network.ip_addrs:
            - interface: eth0
          - cloud.create:
            - names:
              - test-isbm-1
              - test-isbm-2
            - ssh_username: sles
            - image: sles12sp2
            - securitygroup: default
            - size: 'c3.large'
            - location: ap-northeast-1
            - delvol_on_destroy: True


    :return:
    """
    # Detect if this call is using legacy or new style syntax.
    legacy_run = False

    keys = list(kwargs)
    ignored_kwargs = ["name", "__reqs__", "sfun"]
    for item in ignored_kwargs:
        if item in keys:
            keys.remove(item)

    # The rest of the keys should be function names for new-style syntax
    for name in keys:
        if name.find(".") == -1:
            legacy_run = True
    if not keys and kwargs:
        legacy_run = True

    if legacy_run:
        log.debug("Detected legacy module.run syntax: %s", __low__["__id__"])
        return _legacy_run(**kwargs)
    else:
        log.debug("Using new style module.run syntax: %s", __low__["__id__"])
        return _run(**kwargs)


def _run(**kwargs):

    if "name" in kwargs:
        kwargs.pop("name")
    ret = {
        "name": list(kwargs),
        "changes": {},
        "comment": "",
        "result": None,
    }

    functions = [func for func in kwargs if "." in func]
    missing = []
    tests = []
    for func in functions:
        func = func.split(":")[0]
        if func not in __salt__:
            missing.append(func)
        elif __opts__["test"]:
            tests.append(func)

    if not functions:
        ret["comment"] = "No function provided."
        ret["result"] = False
        return ret

    if tests or missing:
        ret["comment"] = " ".join(
            [
                missing
                and "Unavailable function{plr}: {func}.".format(
                    plr=(len(missing) > 1 or ""), func=(", ".join(missing) or "")
                )
                or "",
                tests
                and "Function{plr} {func} to be executed.".format(
                    plr=(len(tests) > 1 or ""), func=", ".join(tests) or ""
                )
                or "",
            ]
        ).strip()

        if missing:
            ret["result"] = False

        return ret

    failures = []
    success = []
    for func in functions:
        _func = func.split(":")[0]
        try:
            func_ret = _call_function(
                _func, returner=kwargs.get("returner"), func_args=kwargs.get(func)
            )
            if not _get_result(func_ret):
                if isinstance(func_ret, dict):
                    failures.append(
                        "'{}' failed: {}".format(
                            func, func_ret.get("comment", "(error message N/A)")
                        )
                    )
                if func_ret is False:
                    failures.append(f"'{func}': {func_ret}")
            else:
                success.append(
                    "{}: {}".format(
                        func,
                        (
                            func_ret.get("comment", "Success")
                            if isinstance(func_ret, dict)
                            else func_ret
                        ),
                    )
                )
                ret["changes"][func] = func_ret
        except (SaltInvocationError, TypeError) as ex:
            failures.append(f"'{func}' failed: {ex}")
    ret["comment"] = ", ".join(failures + success)
    ret["result"] = not bool(failures)

    return ret


def _call_function(name, returner=None, func_args=None, func_kwargs=None):
    """
    Calls a function from the specified module.

    :param str name: module.function of the function to call
    :param dict returner: Returner specification to use.
    :param list func_args: List with args and dicts of kwargs (one dict per kwarg)
        to pass to the function.
    :return: Result of the function call
    """
    if func_args is None:
        func_args = []

    if func_kwargs is None:
        func_kwargs = {}

    mret = salt.utils.functools.call_function(__salt__[name], *func_args, **func_kwargs)
    if returner is not None:
        returners = salt.loader.returners(__opts__, __salt__)
        if returner in returners:
            returners[returner](
                {
                    "id": __opts__["id"],
                    "ret": mret,
                    "fun": name,
                    "jid": salt.utils.jid.gen_jid(__opts__),
                }
            )
    return mret


def _legacy_run(name, **kwargs):
    """
    .. deprecated:: 2017.7.0
       Function name stays the same, behaviour will change.

    Run a single module function

    ``name``
        The module function to execute

    ``returner``
        Specify the returner to send the return of the module execution to

    ``kwargs``
        Pass any arguments needed to execute the function
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": None}
    if name not in __salt__:
        ret["comment"] = f"Module function {name} is not available"
        ret["result"] = False
        return ret

    if __opts__["test"]:
        ret["comment"] = f"Module function {name} is set to execute"
        return ret

    aspec = salt.utils.args.get_function_argspec(__salt__[name])
    args = []
    defaults = {}

    arglen = 0
    deflen = 0
    if isinstance(aspec.args, list):
        arglen = len(aspec.args)
    if isinstance(aspec.defaults, tuple):
        deflen = len(aspec.defaults)
    # Match up the defaults with the respective args
    for ind in range(arglen - 1, -1, -1):
        minus = arglen - ind
        if deflen - minus > -1:
            defaults[aspec.args[ind]] = aspec.defaults[-minus]
    # overwrite passed default kwargs
    for arg in defaults:
        if arg == "name":
            if "m_name" in kwargs:
                defaults[arg] = kwargs.pop("m_name")
        elif arg == "fun":
            if "m_fun" in kwargs:
                defaults[arg] = kwargs.pop("m_fun")
        elif arg == "state":
            if "m_state" in kwargs:
                defaults[arg] = kwargs.pop("m_state")
        elif arg == "saltenv":
            if "m_saltenv" in kwargs:
                defaults[arg] = kwargs.pop("m_saltenv")
        if arg in kwargs:
            defaults[arg] = kwargs.pop(arg)
    missing = set()
    for arg in aspec.args:
        if arg == "name":
            rarg = "m_name"
        elif arg == "fun":
            rarg = "m_fun"
        elif arg == "names":
            rarg = "m_names"
        elif arg == "state":
            rarg = "m_state"
        elif arg == "saltenv":
            rarg = "m_saltenv"
        else:
            rarg = arg
        if rarg not in kwargs and arg not in defaults:
            missing.add(rarg)
            continue
        if arg in defaults:
            args.append(defaults[arg])
        else:
            args.append(kwargs.pop(rarg))
    if missing:
        comment = "The following arguments are missing:"
        for arg in missing:
            comment += f" {arg}"
        ret["comment"] = comment
        ret["result"] = False
        return ret

    if aspec.varargs:
        if aspec.varargs == "name":
            rarg = "m_name"
        elif aspec.varargs == "fun":
            rarg = "m_fun"
        elif aspec.varargs == "names":
            rarg = "m_names"
        elif aspec.varargs == "state":
            rarg = "m_state"
        elif aspec.varargs == "saltenv":
            rarg = "m_saltenv"
        else:
            rarg = aspec.varargs

        if rarg in kwargs:
            varargs = kwargs.pop(rarg)

            if not isinstance(varargs, list):
                msg = "'{0}' must be a list."
                ret["comment"] = msg.format(aspec.varargs)
                ret["result"] = False
                return ret

            args.extend(varargs)

    nkwargs = {}
    if aspec.keywords and aspec.keywords in kwargs:
        nkwargs = kwargs.pop(aspec.keywords)
        if not isinstance(nkwargs, dict):
            msg = "'{0}' must be a dict."
            ret["comment"] = msg.format(aspec.keywords)
            ret["result"] = False
            return ret

    try:
        if aspec.keywords:
            mret = __salt__[name](*args, **nkwargs)
        else:
            mret = __salt__[name](*args)
    except Exception as e:  # pylint: disable=broad-except
        ret["comment"] = "Module function {} threw an exception. Exception: {}".format(
            name, e
        )
        ret["result"] = False
        return ret
    else:
        if mret is not None or mret != {}:
            ret["changes"]["ret"] = mret

    if "returner" in kwargs:
        ret_ret = {
            "id": __opts__["id"],
            "ret": mret,
            "fun": name,
            "jid": salt.utils.jid.gen_jid(__opts__),
        }
        returners = salt.loader.returners(__opts__, __salt__)
        if kwargs["returner"] in returners:
            returners[kwargs["returner"]](ret_ret)
    ret["comment"] = f"Module function {name} executed"
    ret["result"] = _get_result(mret)

    return ret


def _get_result(func_ret):
    res = True
    # if mret a boolean, return that as the result
    if isinstance(func_ret, bool):
        res = func_ret
    # if mret is a dict, check if certain keys exist
    elif isinstance(func_ret, dict):
        res = _get_dict_result(func_ret)

    return res


def _get_dict_result(node):
    ret = True
    for key, val in node.items():
        if (key == "result" and val is False) or (key == "retcode" and val != 0):
            ret = False
            break
        elif isinstance(val, dict):
            ret = _get_dict_result(val)
            if ret is False:
                break
    return ret


mod_watch = salt.utils.functools.alias_function(run, "mod_watch")
